# Load Main Classes into Module here.
# These can be accessed externally (e.g. from tests) using "from model import AirLayer". etc...
from .surface import Surface
from .air import AirLayer
from .radiation import Radiation
from . import radiation
from . import model


var_dict = {
    'T_s' : {
        'symbol' : r"$T_s$",
        'label' : r"$T_s$ (°C)",
        'unit' : r"°C",
        'der' : 'dLE_dTs',
        'tower_col' : 'T_s_CNR4',
    },
    'r_H': {
        'symbol' : r"$r_{H}$",
        'label' : r"$r_{H}$ (s m$^{-1}$)",
        'unit' : r"s m$^{-1}$",
        'der' : 'dLE_drH',
        'tower_col' : 'r_H_CNR4',
    },
    'T_a' : {
        'symbol' : r"$T_a$",
        'label' : r"$T_a$ (°C)",
        'unit' : r"°C",
        'der' : 'dLE_dTa',
        'der_br' : ['dLE_dTa1', 'dLE_dTa2'],
    },
    'h_r' : {
        'symbol' : r"$h_r$",
        'label' : r"$h_r$ (%)",
        'unit' : r"%",
        'der' : 'dLE_dhr',
        'der_br' : ['dLE_dhr1', 'dLE_dhr2'],
    },
    'SW_IN' : {
        'symbol' : r"$R_{\mathrm{SW}}^{\downarrow}$",
        'label' : r"$R_{\mathrm{SW}}^{\downarrow}$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
        'der' : 'dLE_dRsw',
    },
    'p_a' : {
        'symbol' : r"$p_a$",
        'label' : r"$p_a$ (kPa)",
        'unit' : r"kPa",
        'der' : 'dLE_dpa',
        'der_br' : ['dLE_dpa1', 'dLE_dpa2'],
    },    
    'p_s' : {
        'symbol' : r"$p_s$",
        'label' : r"$p_s$ (kPa)",
        'unit' : r"kPa",
        'der' : 'dLE_dps',
    },    
    'T_a1' : {
        'symbol' : r"$T_{a1}$",
        'label' : r"$T_{a1}$ (°C)",
        'unit' : r"°C",
        'der' : 'dLE_dTa1',
    },
    'T_a2' : {
        'symbol' : r"$T_{a2}$",
        'label' : r"$T_{a2}$ (°C)",
        'unit' : r"°C",
        'der' : 'dLE_dTa2',
    },
    'h_r1' : {
        'symbol' : r"$h_{r1}$",
        'label' : r"$h_{r1}$ (%)",
        'unit' : r"%",
        'der' : 'dLE_dhr1',
    },
    'h_r2' : {
        'symbol' : r"$h_{r2}$",
        'label' : r"$h_{r2}$ (%)",
        'unit' : r"%",
        'der' : 'dLE_dhr2',
    },
    'p_a1' : {
        'symbol' : r"$p_{a1}$",
        'label' : r"$p_{a1}$ (kPa)",
        'unit' : r"kPa",
        'der' : 'dLE_dpa1',
    },   
    'p_a2' : {
        'symbol' : r"$p_{a2}$",
        'label' : r"$p_{a2}$ (kPa)",
        'unit' : r"kPa",
        'der' : 'dLE_dpa2',
    },     
    'LE' : {
        'symbol' : r"$\lambda E$",
        'label' : r"$\lambda E$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
    },
    'H' : {
        'symbol' : r"$H$",
        'label' : r"$H$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
    },
    'dLE_dTs' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial T_s}$",
        'label' : r"$\frac{\partial \lambda E}{\partial T_s}$ (W m$^{-2}$ °C$^{-1}$)"
    },
    'dLE_drH' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial r_{H}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial r_{H}}$ (W m$^{-2}$ s$^{-1}$ m)"
    },
    'dLE_dTa' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial T_a}$",
        'label' : r"$\frac{\partial \lambda E}{\partial T_a}$ (W m$^{-2}$ °C$^{-1}$)"
    },
    'dLE_dhr' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial h_r}$",
        'label' : r"$\frac{\partial \lambda E}{\partial h_r}$ (W m$^{-2}$ %$^{-1}$)"
    },
    'dLE_dRsw' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial R_{\mathrm{SW}}^{\downarrow}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial R_{\mathrm{SW}}^{\downarrow}}$ (W m$^{-2}$ W$^{-1}$ m$^{2}$)"
    },
    'dLE_dpa' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial p_a}$",
        'label' : r"$\frac{\partial \lambda E}{\partial p_a}$ (W m$^{-2}$ kPa$^{-1}$)"
    },
    'dLE_dps' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial p_s}$",
        'label' : r"$\frac{\partial \lambda E}{\partial p_s}$ (W m$^{-2}$ kPa$^{-1}$)"
    },
    'dLE_dTa1' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial T_{a1}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial T_{a1}}$ (W m$^{-2}$ °C$^{-1}$)"
    },
    'dLE_dTa2' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial T_{a2}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial T_{a2}}$ (W m$^{-2}$ °C$^{-1}$)"
    },
    'dLE_dhr1' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial h_{r1}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial h_{r1}}$ (W m$^{-2}$ %$^{-1}$)"
    },
    'dLE_dhr2' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial h_{r2}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial h_{r2}}$ (W m$^{-2}$ %$^{-1}$)"
    },
    'dLE_dpa1' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial p_{a1}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial p_{a1}}$ (W m$^{-2}$ kPa$^{-1}$)"
    },
    'dLE_dpa2' : {
        'symbol' : r"$\frac{\partial \lambda E}{\partial p_{a2}}$",
        'label' : r"$\frac{\partial \lambda E}{\partial p_{a2}}$ (W m$^{-2}$ kPa$^{-1}$)"
    },
    'u' : {
        'symbol' : r"$u$",
        'label' : r"$u$ (m s$^{-1}$)",
        'unit' : r"m s$^{-1}$",
    },
    'LW_IN' : {
        'symbol' : r"$R_{\mathrm{LW}}^{\downarrow}$",
        'label' : r"$R_{\mathrm{LW}}^{\downarrow}$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
        'der' : 'dLE_dRsw',
    },    
    'SW_OUT' : {
        'symbol' : r"$R_{\mathrm{SW}}^{\uparrow}$",
        'label' : r"$R_{\mathrm{SW}}^{\uparrow}$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
    },    
    'LW_OUT' : {
        'symbol' : r"$R_{\mathrm{LW}}^{\uparrow}$",
        'label' : r"$R_{\mathrm{LW}}^{\uparrow}$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
    },    
    'R_n' : {
        'symbol' : r"$R_n$",
        'label' : r"$R_n$ (W m$^{-2}$)",
        'unit' : r"W m$^{-2}$",
    },  
    'u_star' : {
        'symbol' : r"$u_{*}$",
        'label' : r"$u_{*}$ (m s$^{-1}$)",
        'unit' : r"m s$^{-1}$",
        'tower_col' : 'ustar',
    },
}



import numpy as np
# import xarray as xr

KAPPA = 0.41
c_pd = 1005.7               # specific heat capacity of dry air at constant pressure [J kg-1 K-1]
R_D = 287.058	 	        # dry gas constant (J kg-1 K-1)
G = 9.8

def calc_r_H(rho_a, c_p, p_a, T_s, T_a, H):
    """
    Calculate total resistance by inverting the sensible heat flux equation.
    """

    theta_coeff = ( (100 / p_a) ** (R_D/c_pd) )
    r_H = rho_a * c_p * ((theta_coeff * (T_s - T_a) ) / H)

    return r_H

def calc_r_aH(surf, u, Psi_M, Psi_H, Psi_M0=0, Psi_H0=0, z=3.8735):
# def calc_r_a(u, z, d_0, z_0m, z_0h, Psi_M, Psi_H):
    """
    Calculate aerodynamic resistance according to MOST.
    """

    # Calculate aerodynamic resistance
    r_a = ( (np.log((z - surf.d_0) / surf.z_0m) - Psi_M + Psi_M0) * \
            (np.log((z - surf.d_0) / surf.z_0h) - Psi_H + Psi_H0) ) / \
                (KAPPA**2 * u)

    return r_a

def calc_r_bH(u_0m, w_l, lai, c):
    """
    Calculate boundary layer resistance.
    """

    r_bH = (c / lai) * ( (w_l / u_0m) ** (1/2) )

    return r_bH

def calc_r_aM(u, u_star):

    r_aM = u / (u_star**2)

    return r_aM

# def calc_u(z, u_star, h, Psi_M, Psi_M0=0):
#     """
#     Calculate wind speed according to the logarithmic wind profile.
#     """
#     d_0 = calc_d0(h)
#     z_0m = calc_z0m(h)
    
#     u = (u_star / KAPPA) * (np.log((z - d_0) / z_0m) - Psi_M + Psi_M0)

#     return u

# def calc_ustar(u, h, Psi_M, Psi_M0=0, z=3.8735):
#     """
#     Calculate friction velocity according to MOST.
#     """
#     d_0 = calc_d0(h)
#     z_0m = calc_z0m(h)

#     u_star = ( u * KAPPA) /  (np.log((z - d_0) / z_0m) - Psi_M + Psi_M0)

#     return u_star


def calc_Psi_M(zeta):
    """
    Calculates the stability correction function for momentum transport.

    Reference: Brutsaert, W. (2005). Hydrology: An Introduction. Cambridge University Press.
    """

    Psi_m = np.real_if_close(np.where(zeta < 0, _calc_Psi_M_unstable(zeta), _calc_Psi_stable(zeta)))

    return Psi_m

def calc_Psi_H(zeta):
    """
    Calculates the stability correction function for heat (and vapour) transport.

    Reference: Brutsaert, W. (2005). Hydrology: An Introduction. Cambridge University Press.
    """
    Psi_h = np.real_if_close(np.where(zeta < 0, _calc_Psi_H_unstable(zeta), _calc_Psi_stable(zeta)))

    return Psi_h

def _calc_Psi_stable(zeta):
    """

    """
    a = 6.1
    b = 2.5
    Psi_m = -a * np.log(zeta + (1.0 + zeta**b)**(1.0/b))    # Eq. (2.59)

    return Psi_m

def _calc_Psi_M_unstable(zeta):
    """
    Calculate Psi_M for unstable conditions (zeta < 0).
    """
    a = 0.33
    b = 0.41
    y = -1 * zeta
    x = (y / a)**(1/3)

    Psi_0 = -np.log(a) + np.sqrt(3) * b * a**(1/3) * (np.pi/6.0)
    # y = min(y, b**(-1/3))
    y = np.where(y < b**(-1/3), y, b**(-1/3))

    Psi_m = np.log(a+y) - 3 * b * y**(1/3) + \
                ((b*a**(1/3)) / 2 ) * np.log( (1+x)**2 / (1 - x + x**2)) + \
                np.sqrt(3) * b * a**(1/3) * np.arctan((2*x - 1) / np.sqrt(3) ) + \
                Psi_0                       # Eq. (2.63)

    return Psi_m

def _calc_Psi_H_unstable(zeta):
    """
    Calculate Psi_H for unstable conditions (zeta < 0).
    """
    c = 0.33
    d = 0.057
    n = 0.78
    y = -1 * zeta

    Psi_h = ((1 - d) / n) * np.log((c + y**n) / c )     # Eq. (2.64)

    return Psi_h  


def calc_Hv(air, H, LE):
# def calc_Hv(H, LE, T_a, p_a, e_a):
    """
    Calculate virtual sensible heat flux.
    """
    # theta = met.calc_theta(T_a, p_a)
    E = LE / air.lambda_v
    # c_p = met.calc_cp(p_a, e_a)

    H_v = H + ( 0.61 * air.c_p * air.theta_a * E )

    return H_v

def calc_L(air, u_star, H, LE):
# def calc_L(self, u_star, theta_v, rho_a, c_p, H_v):
    """
    Calculate the Monin-Obukhov length, L [m].
    """

    H_v = calc_Hv(air, H, LE)
    # theta_v = met.calc_theta(T_a, p_a)

    L = - ( ( u_star**3 ) * air.theta_v ) / ( KAPPA * G * ( H_v / ( air.rho_a * air.c_p ) ) )

    return L

def calc_zeta(z, d_0, L):

    return (z - d_0) / L

def calc_u(z, u_star, d_0, z_0m, Psi_M, Psi_M0=0):
    """
    Calculate wind speed according to the logarithmic wind profile.
    """    
    u = (u_star / KAPPA) * (np.log((z - d_0) / z_0m) - Psi_M + Psi_M0)

    u = np.where(u < 0., 1e-6, u)

    return u

def calc_ustar(u, d_0, z_0m, Psi_M, Psi_M0=0, z=3.8735):
    """
    Calculate friction velocity according to MOST.
    """
    u_star = ( u * KAPPA) /  (np.log((z - d_0) / z_0m) - Psi_M + Psi_M0)

    return u_star


def calc_u0m(u_h, lai, h, d_0, z_0m, w_l):
    """
    Calculate wind speed at the source height for momentum transfer
    """

    a = 0.28 * (lai**(2/3)) * (h**(1/3)) * (w_l**(-1/3))
    # d_0 = res.calc_d0(h)
    # z_0m = res.calc_z0m(h)

    u_0m = u_h * np.exp(-a * (1 - ( (d_0 + z_0m) / h ) ) )

    return u_0m


#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'brynmorgan@ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'GNU General Public License v3.0'
__date__ = 'Wed 24 Aug 22 10:54:04'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = 'https://github.com/brynemorgan/aeroet/blob/main/src/aeroet/air.py'

"""

Name:           air.py
Compatibility:  Python 3.10.0
Description:    Description of what program does

URL:            https://

Requires:       numpy

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        brynmorgan@ucsb.edu
Copyright:      (c) Bryn Morgan 2022

"""


# IMPORTS
import numpy as np

# from model.radiation import calc_emissivity

# CONSTANTS
c_pd = 1005.7               # specific heat capacity of dry air at constant pressure [J kg-1 K-1]
c_pv = 1996.0               # specific heat of water vapour at constant pressure [J kg-1 K-1]
EPSILON = 0.622             # ratio of molar mass of water to molar mass of dry air [unitless]
R_D = 287.058	 	        # dry gas constant (J kg-1 K-1)
R = 8.314                   # ideal gas constant (J mol-1 K-1)
R_v = 461.                  # gas constant for water vapor (J kg-1 K-1)
M_H2O = 0.01801528          # molar mass of water (kg mol-1)
# SIGMA = 5.67036713e-08 		# Stefan-Boltzmann constant [W m-2 K-4]
G = 9.80665                 # gravitational acceleration [m s-2]
M_D = 0.0289635              # molar mass of dry air (kg mol-1)
gamma_d = G / c_pd


class AirLayer:
    """
    
    Parameters
    ----------
    z : float
        Height of the air layer (and met data measurements) [m]
    
    u : float | array-like
        Wind speed [m s-1]
    
    T_a : float | array-like
        Air temperature [K]

    p_a : float | array-like
        Air pressure [kPa]
    
    h_r : float | array-like
        Relative humidity [%]
    
    e_a : float | array-like
        Vapor pressure of air [kPa]
    
    q : float | array-like
        Specific humidity of air [kg kg-1]
    
    rho_a : float | array-like
        Density of air [kg m-3]
    
    c_p : float | array-like
        Heat capacity of air at constant pressure [J kg-1 K-1]
    
    theta_a : float | array-like
        Potential temperature of air [K]
    
    theta_v : float | array-like
        Virtual potential temperature of air [K]


    Methods
    -------
    calc_vp()
        Calculates the vapour pressure of air based on relative humidity and the
        Tetens equation for saturation vapour pressure.

    calc_rho()
        Calculates the density of air based on the ideal gas law.
    
    calc_c_p()
        Calculates heat capacity of (moist) air at constant pressure.

    calc_q()
        Calculate the specific humidity.
    
    calc_theta()
        Calculate the potential temperature of a parcel of air.
    
    calc_Tv()
        Calculate the virtual temperature.

    calc_r_v(self):
        Calculate the mixing ratio of a parcel of air.
    
    calc_slope_svp(self):
        Calculates the slope of the saturation vapour pressure line.
    
    calc_svp()
        Calculates the saturation vapour pressure using the Tetens equation.

    calc_theta_v()
        Calculate virtual potential temperature.

    calc_emissivity()
        Estimates the atmospheric emissivity for a clear sky.

    """

    def __init__(
        self,
        z,
        u,
        T_a,
        p_a,
        h_r,
        c = 0.,
        z_0 = 0.,
        e_a = None,
        c_p = None,
        rho_a = None
    ):
        """
        Parameters
        ----------
        z : float
            Height of the `AirLayer` (and met data measurements) [m]
        u : float | array-like
            Wind speed [m s-1]
        T_a : float | array-like
            Air temperature [K]
        p_a : float | array-like
            Air pressure [kPa]
        h_r : float | array-like
            Relative humidity [%]
        e_a : float | array-like, optional
            Vapor pressure of air [kPa]
        c_p : float | array-like, optional
            Heat capacity of air [J kg-1 K-1], by default None
        rho_a : float | array-like, optional
            Density of air [kg m-3], by default None
        """

        self.z = z
        self.u = u
        self.T_a = T_a
        self.p_a = p_a
        self.h_r = h_r

        # self.z = np.array(z)
        # self.u = np.array(u)
        # self.T_a = np.array(T_a)
        # self.p_a = np.array(p_a)
        # self.h_r = np.array(RH)

        # Vapour pressure, e_a [kPa]
        if e_a is not None:
            self.e_a = e_a
        else:
            self.e_a = self.calc_vp()

        # Specific humidity, q [kg kg-1]
        self.q = self.calc_q()

        # Air density, rho_a [kg m-3]
        if rho_a is not None:
            self.rho_a = rho_a
        else:
            self.rho_a = self.calc_rho()
        
        # Heat capacity, c_p [J kg-1 K-1]
        if c_p is not None:
            self.c_p = c_p
        else:
            self.c_p = self.calc_cp()

        # Latent heat of vaporization, lambda_v [J kg-1]
        self.lambda_v = self.calc_lhvap()

        # Potential temperature, theta_a [K]
        self.theta_a = self.calc_theta(T_a=self.T_a, z_0=z_0, gamma=gamma_d)

        # Virtual potential temperature, theta_v [K]
        self.theta_v = self.calc_theta_v()

        # VPD [kPa]
        self.vpd = self.calc_vpd()

        # Emissivity
        self._cloud_fraction = c
        self.emissivity = self.calc_emissivity()

    def describe(self):
        return self.__dict__.copy()



    def calc_vp(self):
        """
        Calculates the vapour pressure of air based on relative humidity and the
        saturation vapour pressure.

        Parameters
        ----------
        h_r : float
            Relative humidity [%].
        
        e_star : float
            Saturation vapour pressure [kPa].

        Returns
        -------
        e_a : float
            Vapour pressure of air [kPa]

        Notes
        -----
        The vapour pressure of air, |e_a|, is the product of relative 
        humidity, |h_r|, and saturation vapour pressure, |e_star|.
        
        .. math:: 
        
            e_a = \\frac{h_r}{100} e_a^*
        
        .. |e_a| replace:: :math:`e_a` [kPa]
        .. |h_r| replace:: :math:`h_r` [%]
        .. |e_star| replace:: :math:`e_a^*` [kPa]

        """

        e_a = (self.h_r/100) * self.calc_svp()

        return e_a
    
    def calc_rho(self):
        """
        Calculates the density of air based on the ideal gas law.

        Parameters
        ----------
        p_a : float
            Total pressure of air [kPa].
        
        e_a : float
            Vapor pressure of air [kPa].
        
        T_a : float
            Temperature of air [K]

        Returns
        -------
        rho : float
            Density of air [kg m-3]
        
        Notes
        -----
        The density of air, |rho_a|, varies with temperature, 
        pressure, and humidity, and can be calculated from the ideal gas law:

        .. math::
            
            \\rho_a = \\frac{1000 \\, p_a}{R_d T_a} \left( 1 - \\frac{(1 - \epsilon) e_a}{1000 \\, p_a} \\right)
        
        where |R_d [J mol-1 K-1]| is the gas constant for dry air, |T_a [K]| is 
        the temperature of air, |p_a [kPa]| is the total pressure of air, |e_a [kPa]|
        is the vapor pressure of air, and |epsilon| is the ratio of the molar mass
        of water to that of dry air.

        .. |rho_a| replace:: :math:`\\rho_a`  [kg m\ :sup:`-3`]
        .. |R_d [J mol-1 K-1]| replace:: :math:`R_d = 287.058`  J mol\ :sup:`-1` K\ :sup:`-1`
        .. |T_a [K]| replace:: :math:`T_a`  [K]
        .. |p_a [kPa]| replace:: :math:`p_a`  [kPa]
        .. |e_a [kPa]| replace:: :math:`e_a`  [kPa]
        .. |epsilon| replace:: :math:`\epsilon = 0.622`

        """

        rho = ((self.p_a*1000)/(R_D*self.T_a))*(1-(((1-EPSILON)*self.e_a)/(self.p_a*1000)))

        return rho

    def calc_cp(self):
        """
        Calculates heat capacity of (moist) air at constant pressure.

        Parameters
        ----------
        p_a : float
            Total pressure of air. Units must match `e_a`.
        e_a : float
            Vapor pressure of air. Units must match `p_a`.

        Returns
        -------
        c_p : float
            Heat capacity of air at constant pressure |[J kg-1 K-1]|.
        
        Notes
        -----
        The heat capacity of air at constant pressure, |c_p|, is calculated as 
        the sum of the heat capacities of dry air, |c_pd|, and water vapour,
        |c_pv|, at constant pressure, weighted by the specific humidity, |q|:

        .. math::

            c_p = (1 - q) c_{pd} + q c_{pv}
        
        where |c_pd = 1005.7| and |c_pv = 1996.0|.

        .. |[J kg-1 K-1]| replace:: [J kg\ :sup:`-1` K\ :sup:`-1`
        .. |c_p| replace:: :math:`c_p` [J kg\ :sup:`-1` K\ :sup:`-1`]
        .. |c_pd| replace:: :math:`c_{pd}` [J kg\ :sup:`-1` K\ :sup:`-1`]
        .. |c_pv| replace:: :math:`c_{pv}` [J kg\ :sup:`-1` K\ :sup:`-1`]
        .. |q| replace:: :math:`q` [kg kg\ :sup:`-1`]
        .. |c_pd = 1005.7| replace:: :math:`c_{pd} = 1005.7` J kg\ :sup:`-1` K\ :sup:`-1`
        .. |c_pv = 1996.0| replace:: :math:`c_{pv} = 1996.0` J kg\ :sup:`-1` K\ :sup:`-1`
        """
        # Calculate specific humidity
        # q = self.calc_q(self.p_a, self.e_a)w

        # Calculate heat capacity of moist air
        c_p = (1 -self.q) * c_pd + self.q * c_pv

        return c_p

    def calc_lhvap(self):
        """
        Calculates the latent heat of vaporization |[J kg-1]| of water at a given
        temperature.

        Parameters
        ----------
        T_K : float
            Temperature [K].

        Returns
        -------
        lambda_v : float
            Latent heat of vaporization |[J kg-1]|

        Notes
        -----
        The latent heat of vaporization, |lambda_v|, is calculated
        as a function of temperature, |T|, using the following equation [1]_:
       
        .. math::
            
            \lambda_v = (2.501 - 2.361 \\times 10^{-3} T) \\times 10^6    
        
        .. |[J kg-1]| replace:: [J kg\ :sup:`-1`]
        .. |lambda_v| replace:: :math:`\lambda_v` [J kg-1]    
        .. |T| replace:: :math:`T` [K]
        
        References
        ----------
        .. [1] Allen, R. G., Pereira, L. S., Raes, D., & Smith, M. (1998). Crop 
            evapotranspiration: Guidelines for computing crop water requirements 
            (FAO Irrigation and Drainage Paper 56). Food and Agriculture Organization 
            of the United Nations. https://www.fao.org/3/x0490e/x0490e00.htm. (Eq. 3-1).
        """
        # Calculate latent heat of vaporization at T
        lambda_v = (2.501 - (2.361e-3 * (self.T_a - 273.15))) * 1e6

        return lambda_v
    
    def calc_q(self):
        """
        Calculate the specific humidity of air.

        Parameters
        ----------
        p_a : float
            Total pressure of air. Units must match `e_a`.
        e_a : float
            Vapor pressure of air. Units must match `p_a`.

        Returns
        -------
        q : float
            Specific humidity |[kg kg-1]|.

        Notes
        -----
        The specific humidity, |q|, is calculated from the vapour pressure |e_a| 
        and air pressure |p_a| according to Cambell and Norman (1998), Eq. 3.19 [1]_:

        .. math::

            q = \\frac{\epsilon \\, e_a}{(p_a - e_a (1 - \epsilon))}
        
        where |epsilon| is the ratio of the molecular weight of water vapour to
        that of dry air.
        
        
        .. |[kg kg-1]| replace:: [kg kg\ :sup:`-1`]
        .. |q| replace:: :math:`q` [kg kg\ :sup:`-1`]
        .. |e_a| replace:: :math:`e_a` [kPa]
        .. |p_a| replace:: :math:`p_a` [kPa]
        .. |epsilon| replace:: :math:`\epsilon = 0.622`
        

        References
        ----------
        .. [1] Campbell, G. S., & Norman, J. (1998). An Introduction to Environmental 
            Biophysics. Springer Science & Business Media.
        """
        q = (EPSILON * self.e_a) / (self.p_a - (1-EPSILON) * self.e_a)

        return q


    def calc_theta(self, T_a, z_0=0.0, gamma=gamma_d, p_0=100.):
        """
        Calculate the potential temperature of a parcel of air.

        Parameters
        ----------
        T_a : float
            Air temperature [K]
        p_a : float
            Air pressure [kPa]
        p_0 : float
            Reference air pressure [kPa]. The default is 100.0.

        Returns
        -------
        theta : float
            Potential temperature [K].
        
        Notes
        -----
        The potential temperature, |theta|, of air at some reference height |z| 
        above the surface, |z_0|, is given by the equation [1]_:

        .. math::

            \\theta_a = T_a + (z - z_0) \Gamma_d
        
        where |Gamma_d| is the dry adiabatic lapse rate,  which is a function of 
        the gravitational constant, |g|, and the heat capacity of dry air at 
        constant pressure, |c_pd|.

        .. |theta| replace:: :math:`\\theta_a` [K]
        .. |z| replace:: :math:`z` [m]
        .. |z_0| replace:: :math:`z_0` [m]
        .. |Gamma_d| replace:: :math:`\Gamma_d` [K m\ :sup:`-1`]
        .. |g| replace:: :math:`g = 9.8` m s\ :sup:`-2`
        .. |c_pd| replace:: :math:`c_{pd} = 1005.7` J kg\ :sup:`-1` K\ :sup:`-1`

        References
        ----------
        .. [1] Bonan, G. B. (2016). Ecological Climatology: Concepts and Applications. 
            Cambridge University Press.
        """
        # NOTE: This is going to be a problem for taller canopies... See Notes 
        # from Wed, 10 May 2023.
        # https://glossary.ametsoc.org/wiki/Potential_temperature
        # theta = T_a * ( (p_0 / self.p_a) ** (R_D/c_pd) )
        theta = T_a + (self.z - z_0) * gamma

        return theta

    def calc_vpd(self):
        """
        Calculate the vapour pressure deficit.

        Returns
        -------
        vpd : float
            Vapour pressure deficit of air [kPa].
        
        Notes
        -----
        The vapour pressure deficit, |VPD|, is defined as the difference
        between the saturation vapour pressure, |e_star|, and the actual
        vapour pressure, |e_a|:

        .. math::

            \\text{VPD} = e_a^* - e_a.
        
        .. |VPD| replace:: :math:`\\text{VPD}` [kPa]
        .. |e_a| replace:: :math:`e_a` [kPa]
        .. |e_star| replace:: :math:`e_a^*` [kPa]
        """

        vpd = self.calc_svp() - self.e_a

        return vpd

    def calc_lapse_rate(self):
        """
        Calculate the lapse rate of air.

        Returns
        -------
        gamma : float
            Lapse rate of air |[K m-1]|
        
        Notes
        -----
        The lapse rate, |gamma|, is defined as the rate of change of temperature 
        with height, |z|:

        .. math::

            \Gamma = - \\frac{dT_a}{dz}
        
        .. |[K m-1]| replace:: [K m\ :sup:`-1`]
        .. |gamma| replace:: :math:`\Gamma` [K m\ :sup:`-1`]
        .. |z| replace:: :math:`z` [m]
        """
        
        r_v = self.calc_r_v()

        gamma = G * ( 
            (1 + (self.lambda_v * r_v) / (R_D * self.T_a)) / 
            ( c_pd + ((self.lambda_v**2 * r_v) / (R_v * self.T_a**2)) ) 
        )
        return gamma


    def calc_svp(self):
        """
        Calculates the saturation vapour pressure using the Tetens equation.

        Parameters
        ----------
        T_a : float
            Temperature [K]

        Returns
        -------
        e_star : float
            Saturation vapour pressure [kPa]

        Notes
        -----
        The saturation vapour pressure, |e_star|, is calculated using the Tetens
        equation [1]_:

        .. math::

            e_a^* = a \exp \left( \\frac{b (T_a+273.15)}{c + T_a + 273.15} \\right)
        
        where |a = 0.611|, |b = 17.502|, and |c = 240.97| are 
        constants.

        .. |e_star| replace:: :math:`e_a^*` [kPa]
        .. |a = 0.611| replace:: :math:`a = 0.611`
        .. |b = 17.502| replace:: :math:`b = 17.502`
        .. |c = 240.97| replace:: :math:`c = 240.97`

        References
        ----------
        .. [1] Buck, A. L. (1981). New Equations for Computing Vapor Pressure and 
           Enhancement Factor. Journal of Applied Meteorology and Climatology, 
           20(12). https://doi.org/10.1175/1520-0450(1981)020%3C1527:NEFCVP%3E2.0.CO;2


        """
        # TODO: Raise a warning if T_a is not in range for Teten's equation.
        T_C = self.T_a - 273.15

        e_star = 0.611 * np.exp((17.502 * T_C) / (T_C + 240.97))

        return e_star


    def calc_slope_svp(self):
        """
        Calculates the slope of the saturation vapour pressure line.

        Parameters
        ----------
        T_a : float
            Temperature [K]

        Returns
        -------
        delta : float
            Slope of the saturation vapour pressure curve [unitless]

        Notes
        -----
        The slope of the saturation vapour pressure curve, |delta|, is calculated as [1]_:

        .. math::

            \delta = \\frac{b \cdot c \cdot e_a^*}{(c + T_a + 273.15)^2}
        
        where |e_star| is the saturation vapour pressure, |T_a| is the air temperature,
        |a = 0.611|, |b = 17.502|, and |c = 240.97|.

        .. |delta| replace:: :math:`\delta`
        .. |e_star| replace:: :math:`e_a^*` [kPa]
        .. |T_a| replace:: :math:`T_a` [K]
        .. |a = 0.611| replace:: :math:`a = 0.611`
        .. |b = 17.502| replace:: :math:`b = 17.502`
        .. |c = 240.97| replace:: :math:`c = 240.97`

        References
        ----------
        .. [1] Campbell, G. S., & Norman, J. (1998). An Introduction to Environmental 
            Biophysics. Springer Science & Business Media. (Eq. 3.9).
        """
        T_C = self.T_a - 273.15
        e_star = self.calc_svp()

        delta = (17.502 * 240.97 * e_star) / ((240.97 + T_C)**2)

        return delta
    


    def calc_Tv(self):
        """
        Calculate the virtual temperature.

        Parameters
        ----------
        T_a : float
            Air temperature [K]
        
        p_a : float
            Air pressure [kPa]
        
        e_a : float
            Vapour pressure [kPa]

        Returns
        -------
        T_v : float
            Virtual air temperature [K]
        
        Notes
        -----
        The virtual temperature, |T_v|, is calculated as [4]_:

        .. math::

            T_v = T_a \left( \\frac{1 + r_v}{\epsilon (1 + r_v)} \\right)
        
        where |r_v| is the mixing ratio, |T_a| is the air temperature, and |epsilon| 
        is the ratio of the gas constants for dry air and water vapour.

        .. |T_v| replace:: :math:`T_v` [K]
        .. |r_v| replace:: :math:`r_v` [--]
        .. |T_a| replace:: :math:`T_a` [K]

        References
        ----------
        .. [4] https://glossary.ametsoc.org/wiki/Virtual_temperature
        """

        r_v = self.calc_r_v()

        T_v = self.T_a * ( (1 + r_v) / (EPSILON * (1 + r_v)) )

        return T_v
    

    def calc_r_v(self):
        """
        Calculate the mixing ratio of a parcel of air.

        Parameters
        ----------
        p_a : float
            Total pressure of air. Units must match `e_a`.
        
        e_a : float
            Vapour pressure of air. Units must match `p_a`.
        
        Returns
        -------
        r_v : float
            Mixing ratio (unitless)
        
        Notes
        -----
        The mixing ratio, |r_v|, is calculated as:

        .. math::

            r_v = \\frac{\epsilon e_a}{p_a - e_a}
        
        where |e_a| is the vapour pressure of air, |p_a| is the total pressure of air,
        and |epsilon| is the ratio of the gas constants for dry air and water vapour.

        .. |r_v| replace:: :math:`r_v` [unitless]
        .. |e_a| replace:: :math:`e_a` [kPa]
        .. |p_a| replace:: :math:`p_a` [kPa]
        .. |epsilon| replace:: :math:`\epsilon = 0.622`
            
        """

        r_v = (EPSILON * self.e_a) / (self.p_a - self.e_a)

        return r_v
    

    def calc_theta_v(self):
        """
        Calculate virtual potential temperature.

        Parameters
        ----------
        theta : float
            Potential temperature [K].

        q : float
            Specific humidity [kg kg-1].

        Returns
        -------
        theta_v : float
            Virtual potential temperature [K]
        
        Notes
        -----
        The virtual potential temperature, |theta_v|, is calculated as:

        .. math::

            \\theta_v = \\theta_a (1 + 0.61 q)
        
        where |theta| is the potential temperature, and |q| is the specific humidity [1]_.

        .. |theta_v| replace:: :math:`\\theta_v` [K]
        .. |theta| replace:: :math:`\\theta_a` [K]
        .. |q| replace:: :math:`q` [kg kg\ :sup:`-1`]

        References 
        ----------
        .. [1] https://glossary.ametsoc.org/wiki/Virtual_potential_temperature
        """
        theta_v = self.theta_a * (1 + 0.61*self.q)

        return theta_v

    def calc_psychrometric(self):
        """
        Calculate the psychrometric constant.

        Parameters
        ----------
        c_p : float
            Specific heat of air at constant pressure [J kg-1 K-1]
        
        p_a : float
            Air pressure [kPa]
        
        lambda_v : float
            Latent heat of vaporization [J kg-1]
        
        Returns
        -------
        gamma : float
            Psychrometric constant [kPa K-1]
        
        Notes
        -----
        The psychrometric constant, |gamma|, is calculated as [1]_:

        .. math::

            \gamma = \\frac{c_p p_a}{\\lambda_v \\epsilon}
        
        where |c_p| is the specific heat of air at constant pressure, |p_a| is the air
        pressure, |lambda_v| is the latent heat of vaporization, and |epsilon| is the
        ratio of the gas constants for dry air and water vapour (0.622).
        
        .. |gamma| replace:: :math:`\gamma` [kPa K\ :sup:`-1`]
        .. |c_p| replace:: :math:`c_p` [J kg\ :sup:`-1` K\ :sup:`-1`]
        .. |p_a| replace:: :math:`p_a` [kPa]
        .. |lambda_v| replace:: :math:`\lambda_v` [J kg\ :sup:`-1`]
        .. |epsilon| replace:: :math:`\epsilon` [--]

        References
        ----------

    
        """
        gamma = (self.c_p * self.p_a) / (self.lambda_v * EPSILON)

        return gamma


    
    def calc_emissivity(self):
    # def calc_emissivity(e_a,T_a):
        """
        Estimates the atmospheric emissivity for a clear sky.

        Parameters
        ----------
        e_a : float
            Vapor pressure of air [kPa].

        T_a : float
            Temperature of air [K]
        
        c : float, optional
            Cloud cover fraction [--].

        Returns
        -------
        epsilon_a : float
            Atmospheric emissivity [--]
        
        Notes
        -----
        Atmospheric emissivity, |epsilon_a|, is calculated from the vapor 
        pressure, |e_a|, and temperature of air, |T_a|, as [1]_:

        .. math::

            \epsilon_a = 1.24 \left( \\frac{e_a}{T_a} \\right)^{1/7}
        
        .. |epsilon_a| replace:: :math:`\epsilon_a` [--]
        .. |e_a| replace:: :math:`e_a` [kPa]
        .. |T_a| replace:: :math:`T_a` [K]

        References
        ----------
        .. [1] Brutsaert, W. (1975) On a derivable formula for long-wave radiation
            from clear skies, Water Resour. Res., 11(5), 742-744,
            htpp://dx.doi.org/10.1029/WR011i005p00742.
        """

        epsilon_ac = 1.24 * ((self.e_a * 10) / self.T_a )**(1./7.)

        epsilon_a = self._calc_cloudy_emissivity(epsilon_ac, self._cloud_fraction)

        return epsilon_a



    def _calc_cloudy_emissivity(self, eps_ac, c=0):
        """
        Calculates the atmospheric emissivity for a cloudy sky. Note that if c = 0, 
        then eps_a = eps_ac.

        Parameters
        ----------
        eps_ac : float
            Clear-sky atmospheric emissivity [-].
        
        c : float
            Cloud cover fraction [--]. The default is 0 (clear sky).

        Returns
        -------
        eps_a : float
            Cloudy-sky atmospheric emissivity [-].

        Reference: Campbell and Norman (1998), Eq. 10.12
        Campbell, G. S. and J. M. Norman (1998), An Introduction to Environmental
            Biophysics, 2nd Ed., Springer, New York, 286 pp.

        NOTE: Based on tower data, this formula is not great.
        """

        eps_a = ( (1 - 0.84 * c) * eps_ac ) + (0.84 * c)

        return eps_a




    def calc_wvc(self):
        """
        Calculate the water vapour content of air [mm Hg]

        Parameters
        ----------
        h_r : float | array-like
            Relative humidity [%]
        
        T_a : float | array_like
            Air temperature [K]

        Returns
        -------
        wvc : float | array-like
            Water vapour concentration [mm Hg]
        
        Notes
        -----
        The water vapour content of air, |wvc [mm Hg]|, is calculated as [1]_, [2]_:

        .. math::

            \omega = \\frac{h_r}{100} \cdot e^{( h_1 T_a^3 + h_2 T_a^2 + h_3 T_a + h_4 )}
        
        where |h_r| is the relative humidity, |T_a| is the air temperature, and
        |h_1|, |h_2|, |h_3|, and |h_4| are constants.

        .. |wvc [mm Hg]| replace:: :math:`\omega` [mm Hg]
        .. |h_r| replace:: :math:`h_r` [\%]
        .. |T_a| replace:: :math:`T_a` [K]
        .. |h_1| replace:: :math:`h_1 = 6.4385 \\times 10^{-7}`
        .. |h_2| replace:: :math:`h_2 = -2.7816 \\times 10^{-4}`
        .. |h_3| replace:: :math:`h_3 = 6.939 \\times 10^{-2}`
        .. |h_4| replace:: :math:`h_4 = 1.5587`

        References
        ----------
        .. [1] Tran, Q. H., Han, D., Kang, C., Haldar, A., & Huh, J. (2017). 
            Effects of Ambient Temperature and Relative Humidity on Subsurface 
            Defect Detection in Concrete Structures by Active Thermal Imaging. 
            Sensors, 17(8), Article 8. https://doi.org/10.3390/s17081718

        .. [2] Minkina, W., & Dudzik, S. (2009). Infrared Thermography: 
            Errors and Uncertainties. John Wiley & Sons.
        """
        h1 = 6.8455e-7
        h2 = -2.7816e-4
        h3 = 6.939e-2
        h4 = 1.5587

        T_aC = self.T_a - 273.15

        wvc = (self.h_r/100) * np.exp( 
            h1 * T_aC**3 + h2 * T_aC**2 + h3 * T_aC + h4
        )

        return wvc


    def calc_tau(self, wvc, d):
        """
        Calculate atmospheric transmissivity 

        Parameters
        ----------
        wvc : float | array_like
            Water vapour concentration [mm Hg]
        d : float | array-like
            Distance between surface and sensor [m]

        Returns
        -------
        tau : float | array_like
            Atmospheric transmissivity.

        Notes
        -----
        The atmospheric transmissivity, |tau|, is calculated as [1]_, [2]_:

        .. math::

            \\tau = K_{\\text{atm}} \cdot e^{-d^{1/2} (\\alpha_1 + \\beta_1 \omega ^{1/2}) } + ( 1 -K_{\\text{atm}}) \cdot e^{-d^{1/2} ( x_2 + \\beta_2 \omega ^{1/2} ) }

        where |K_atm| is the atmospheric transmissivity constant, |d| is the
        distance between the surface and the sensor, |alpha_1|, |beta_1|,
        |alpha_2|, and |beta_2| are constants, and |omega| is the water vapour
        concentration.

        .. |tau| replace:: :math:`\tau`
        .. |K_atm| replace:: :math:`K_{\\text{atm}} = 1.9`
        .. |d| replace:: :math:`d` [m]
        .. |alpha_1| replace:: :math:`\\alpha_1 = 6.569 \\times 10^{-3}`
        .. |beta_1| replace:: :math:`\\beta_1 = -2.276 \\times 10^{-3}`
        .. |alpha_2| replace:: :math:`\\alpha_2 = 0.01262`
        .. |beta_2| replace:: :math:`\\beta_2 = -6.67 \\times 10^{-3}`
        .. |omega| replace:: :math:`\omega` [mm Hg]

        References
        ----------
        .. [1] Tran, Q. H., Han, D., Kang, C., Haldar, A., & Huh, J. (2017). 
            Effects of Ambient Temperature and Relative Humidity on Subsurface 
            Defect Detection in Concrete Structures by Active Thermal Imaging. 
            Sensors, 17(8), Article 8. https://doi.org/10.3390/s17081718

        .. [2] Minkina, W., & Klecha, D. (2015). Modeling of Atmospheric Transmission 
            Coefficient in Infrared for Thermovision Measurements. Sensors and IRS² 2015, 
            903–907. https://doi.org/10.5162/irs2015/1.4
        """

        K_atm = 1.9
        a1 = 6.569e-3
        a2 = 0.01262
        b1 = -2.276e-3
        b2 = -6.67e-3

        tau = K_atm * np.exp( -(d**(1/2)) * (a1 + b1 * wvc**(1/2)) ) + \
            (1 - K_atm) * np.exp( -(d**(1/2)) * (a2 + b2 * wvc**(1/2)) )

        return tau


#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2023'

__license__ = 'MIT'
__date__ = 'Wed 11 Jan 23 15:51:50'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           partials.py
Compatibility:  Python 3.7.0
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       None

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2023


"""


# CONSTANTS
c_pd = 1005.7               # specific heat capacity of dry air at constant pressure [J kg-1 K-1]
c_pv = 1996.0               # specific heat of water vapour at constant pressure [J kg-1 K-1]
EPSILON = 0.622             # ratio of molar mass of water to molar mass of dry air [unitless]
R_D = 287.058	 	        # dry gas constant (J kg-1 K-1)
R = 8.314                   # ideal gas constant (J mol-1 K-1)
M_H2O = 0.01801528          # molar mass of water (kg mol-1)
SIGMA = 5.67036713e-08 		# Stefan-Boltzmann constant [W m-2 K-4]

# FUNCTIONS

'''
PROPERTIES OF AN AIR LAYER
z : float
    Height of the air layer (and met data measurements) [m]
u : float | array-like
    Wind speed [m s-1]
T_a : float | array-like
    Air temperature [K]
p_a : float | array-like
    Air pressure [kPa]
RH : float | array-like
    Relative humidity [%]
e_a : float | array-like, optional
    Vapor pressure of air [kPa]
c_p : float | array-like, optional
    Heat capacity of air [J kg-1 K-1], by default None
rho_a : float | array-like, optional
    Density of air [kg m-3], by default None
q
theta_a
theta_v
vpd

Can calculate svp, epsilon_a

'''

def calc_dea_dTa(airlayer):
    """
    Calculate the derivative of vapor pressure with respect to air temperature.

    Returns
    -------
    dea_dTa : float
        Derivative of vapor pressure with respect to air temperature |[kPa K-1]|.

    Notes
    -----
    The derivative of vapor pressure (|e_a|) with respect to air temperature (|T_a|)
    is:

    .. math::

        \\frac{\partial e_a}{\partial T_a} = \\frac{17.502 \\cdot 240.97 \\cdot e_a}{(T_a - 273.15 + 240.97)^2}

    .. |[kPa K-1]| replace:: :math: [kPa K\ :sup:`-1`]
    .. |dea_dTa| replace:: :math:`\\frac{\partial e_a}{\partial T_a}` [kPa/K]
    .. |e_a| replace:: :math:`e_a` [kPa]
    .. |T_a| replace:: :math:`T_a` [K]

    See Also
    --------
    calc_dea_dhr
        The derivative of vapor pressure with respect to relative humidity.

    """
    dea_dTa = (17.502 * 240.97 * airlayer.e_a) / ((airlayer.T_a - 273.15 + 240.97) ** 2)
    return dea_dTa

def calc_dea_dhr(airlayer):
    """
    Calculate the derivative of vapor pressure with respect to relative humidity.

    Returns
    -------
    dea_dhr : float
        Derivative of vapor pressure with respect to relative humidity |[kPa %]|.

    Notes
    -----
    This function computes the derivative of vapor pressure (e_a) with respect to relative humidity (hr)
    as a constant value (1/100):

    .. math::

        \\frac{\partial e_a}{\partial h_r} = \\frac{1}{100}

    .. |[kPa %]| replace:: :math: [kPa %\ :sup:`-1`]
    .. |dea_dhr| replace:: :math:`\\frac{\partial e_a}{\partial h_r}` [kPa/%]
    .. |e_a| replace:: :math:`e_a` [kPa]

    See Also
    --------
    calc_dea_dTa
        The derivative of vapor pressure with respect to air temperature.
    """
    dea_dhr = airlayer.calc_svp() / 100.0
    return dea_dhr

def calc_deps_a_dTa(airlayer):
    """
    Calculate the derivative of emissivity of air with respect to air temperature.

    Returns
    -------
    deps_a_dTa : float
        Derivative of emissivity of air with respect to air temperature [-].

    Notes
    -----
    The derivative of the emissivity of air (|eps_a|) with 
    respect to air temperature (|T_a|) is:

    .. math::

        \\frac{{d(\\varepsilon_a)}}{{dT_a}} = \\frac{1}{7} \\varepsilon_a \\left(\\frac{{17.502 \\cdot 240.97}}{{(T_a - 273.15 + 240.97)^2}} - \\frac{1}{{T_a}}\\right)

    .. |deps_a_dTa| replace:: :math:`\\frac{{d(\\varepsilon_a)}}{{dT_a}}` [-]
    .. |eps_a| replace:: :math:`\\varepsilon_a` [-]
    .. |T_a| replace:: :math:`T_a` [K]

    See Also
    --------
    calc_deps_a_dhr
        The derivative of emissivity of air with respect to relative humidity.
    """
    eps_a = airlayer.calc_emissivity()
    deps_a_dTa = (1.0 / 7.0) * eps_a * (
        ((17.502 * 240.97) / ((airlayer.T_a - 273.15 + 240.97) ** 2))
        - (1 / airlayer.T_a)
    )
    return deps_a_dTa

def calc_deps_a_dhr(airlayer):
    """
    Calculate the derivative of emissivity of air with respect to relative humidity.

    Returns
    -------
    deps_a_dhr : float
        Derivative of emissivity of air with respect to relative humidity [-].

    Notes
    -----
    This function computes the derivative of the emissivity of air (|eps_a|) with 
    respect to relative humidity (|h_r|) using the simplified relation:

    .. math::

        \\frac{\\partial \\varepsilon_a}{\\partial dhr}} = \\frac{1.24}{7} \\left(\\frac{{10 \\cdot e_a}}{{T_a}}\\right)^{-6/7} \\frac{{e^*}}{{10 \\cdot T_a}}

    .. |deps_a_dhr| replace:: :math:`\\frac{\\partial \\varepsilon_a }{{d h_r}}` [-]
    .. |eps_a| replace:: :math:`\\varepsilon_a` [-]
    .. |h_r| replace:: :math:`h_r` [%]
    .. |e_a| replace:: :math:`e_a` [kPa]
    .. |T_a| replace:: :math:`T_a` [K]
    .. |e^*| replace:: :math:`e^*` [kPa]
    """
    e_star = airlayer.calc_svp()
    deps_a_dhr = (1.24 / 7.0) * ((10.0 * airlayer.e_a / airlayer.T_a) ** (-6.0 / 7.0)) * (
        e_star / (10.0 * airlayer.T_a)
    )
    return deps_a_dhr




# # VAPOUR PRESSURE
# def calc_dea_dTa(airlayer):

#     dea_dTa = (17.502 * 240.97 * airlayer.e_a) / ((airlayer.T_a - 273.15 + 240.97)**2) 

#     return dea_dTa

# def calc_dea_dhr(airlayer):

#     dea_dhr = airlayer.calc_svp() / 100.

#     return dea_dhr


# # EMISSIVITY OF AIR
# def calc_deps_a_dTa(airlayer):

#     eps_a = airlayer.calc_emissivity()

#     deps_a_dTa = ( (1. / 7.) * eps_a * (
#         ((17.502 * 240.97) / ((airlayer.T_a - 273.15 + 240.97)**2)  ) 
#         - (1 / airlayer.T_a) 
#     ) )

#     return deps_a_dTa

# def calc_deps_a_dhr(airlayer):

#     e_star = airlayer.calc_svp()

#     deps_a_dhr = ( (1.24 / 7.) * 
#         ((10. * airlayer.e_a / airlayer.T_a)**(-6./7.)) * 
#         ( e_star / (10. * airlayer.T_a) ) 
#     )

#     return deps_a_dhr


# SPECIFIC HUMIDITY
def calc_dq_dTa(airlayer):
    
    dea_dTa = calc_dea_dTa(airlayer)
    dq_dTa = ( (EPSILON * airlayer.p_a) / ( (airlayer.p_a - ( airlayer.e_a * (1 - EPSILON) ) )**2 ) ) * dea_dTa

    return dq_dTa

def calc_dq_dhr(airlayer):

    dea_dhr = calc_dea_dhr(airlayer)
    dq_dhr = ( (EPSILON * airlayer.p_a) / ( (airlayer.p_a - ( airlayer.e_a * (1 - EPSILON) ) )**2 ) ) * dea_dhr

    return dq_dhr

def calc_dq_dpa(airlayer):

    dq_dpa =  (-EPSILON * airlayer.e_a) / ( (airlayer.p_a - airlayer.e_a * (1 - EPSILON))**2 )

    return dq_dpa


# HEAT CAPACITY
def calc_dcp_dTa(airlayer):

    dq_dTa = calc_dq_dTa(airlayer)

    dcp_dTa = (c_pv - c_pd) * dq_dTa

    return dcp_dTa

def calc_dcp_dhr(airlayer):

    dq_dhr = calc_dq_dhr(airlayer)

    dcp_dhr = (c_pv - c_pd) * dq_dhr

    return dcp_dhr

def calc_dcp_dpa(airlayer):

    dq_dpa = calc_dq_dpa(airlayer)

    dcp_dpa = (c_pv - c_pd) * dq_dpa

    return dcp_dpa



# AIR DENSITY
def calc_drho_dTa(airlayer):

    dea_dTa = calc_dea_dTa(airlayer)
    
    drho_dTa = - (1 / airlayer.T_a) * ( ((1 - EPSILON) / R_D) * dea_dTa + airlayer.rho_a )

    return drho_dTa

def calc_drho_dhr(airlayer):

    drho_dhr = - ((1 - EPSILON) / (R_D * airlayer.T_a)) * calc_dea_dhr(airlayer)
    
    return drho_dhr

def calc_drho_dpa(airlayer):

    drho_dpa = 1000 / (R_D * airlayer.T_a)

    return drho_dpa


# POTENTIAL TEMPERATURE
def calc_dtheta_dTa(airlayer):
    
    # dtheta_dTa = (100. / airlayer.p_a )**(R_D / c_pd)
    dtheta_dTa = 1

    return dtheta_dTa

def calc_dtheta_dpa(airlayer):

    # dtheta_dpa = - (R_D / c_pd) * (airlayer.theta_a / airlayer.p_a)
    dtheta_dpa = 0.

    return dtheta_dpa

# LATENT HEAT OF VAPORIZATION
def calc_dlhvap_dTa(airlayer=None):
    return -2.361e3

#-------------------------------------------------------------------------------
# RADIATION COMPONENTS
#-------------------------------------------------------------------------------
# LONGWAVE IN
def calc_dLWin_dTa(airlayer, epsilon_s=0.98):

    deps_a_dTa = calc_deps_a_dTa(airlayer)

    dLWin_dTa = (
        SIGMA * epsilon_s * (airlayer.T_a**3) * 
        (4 * airlayer.calc_emissivity() + airlayer.T_a * deps_a_dTa)
    )

    return dLWin_dTa

def calc_dLWin_dhr(airlayer, epsilon_s=0.98):

    deps_a_dhr = calc_deps_a_dhr(airlayer)

    dLWin_dhr = ( SIGMA * epsilon_s * (airlayer.T_a**4) * deps_a_dhr )

    return dLWin_dhr


# LONGWAVE OUT
def calc_dLWout_dTs(T_s, epsilon_s):

    dLWout_dTs = (-4 * SIGMA * epsilon_s * (T_s**3))

    return dLWout_dTs

#-------------------------------------------------------------------------------
# NET RADIATION
#-------------------------------------------------------------------------------

# def calc_dRn_dRsw(albedo):

#     dRn_dRsw = 1 - albedo

#     return dRn_dRsw


#-------------------------------------------------------------------------------
# SENSIBLE HEAT FLUX
#-------------------------------------------------------------------------------
def calc_dH_dTs(airlayer, r_H, p_s=100.):

    # theta_coeff = (100. / p_s) ** (R_D / c_pd)
    dtheta_dTa = calc_dtheta_dTa(airlayer)

    dH_dTs = airlayer.rho_a * airlayer.c_p * dtheta_dTa / r_H

    return dH_dTs

def calc_dH_drH(airlayer, r_H, theta_s):
    
    dH_drH = - airlayer.rho_a * airlayer.c_p * ((theta_s - airlayer.theta_a) / (r_H**2))

    return dH_drH


def calc_dH_dTa(airlayer, r_H, theta_s):

    # theta_coeff = airlayer.theta_a / airlayer.T_a
    dtheta_dTa = calc_dtheta_dTa(airlayer)
    # theta_s = T_s * theta_coeff

    drho_dTa = calc_drho_dTa(airlayer)
    dcp_dTa = calc_dcp_dTa(airlayer)

    dH_dTa = (1 / r_H) * (
        ( ( drho_dTa * airlayer.c_p + dcp_dTa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) ) 
        - airlayer.rho_a * airlayer.c_p * dtheta_dTa
    )

    return dH_dTa


def calc_dH_dhr(airlayer, r_H, theta_s):
    
    # theta_s = T_s * (airlayer.theta_a / airlayer.T_a)

    drho_dhr = calc_drho_dhr(airlayer)
    dcp_dhr = calc_dcp_dhr(airlayer)

    dH_dhr = ((theta_s - airlayer.theta_a) / r_H) * (drho_dhr * airlayer.c_p + dcp_dhr * airlayer.rho_a)

    return dH_dhr


# NOTE: dH_dpa derivatives might change/should change (at the least need to be 
# cleaned up) due to different theta_a calcs


# def calc_dH_dpa(airlayer, r_H, theta_s, p_s):

#     # theta_s = T_s * (airlayer.theta_a / airlayer.T_a)

#     drho_dpa = calc_drho_dpa(airlayer)
#     dcp_dpa = calc_dcp_dpa(airlayer)

#     dtheta_dpa = calc_dtheta_dpa(airlayer)

#     if isinstance(p_s, float):
#         if p_s == airlayer.p_a:
#             dH_dpa = (1 / r_H) * ( 
#                 ( ( drho_dpa * airlayer.c_p + dcp_dpa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) )
#                 + airlayer.rho_a * airlayer.c_p * dtheta_dpa
#             )
#         else:
#             dH_dpa = (1 / r_H) * ( 
#                 ( ( drho_dpa * airlayer.c_p + dcp_dpa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) )
#                 - airlayer.rho_a * airlayer.c_p * dtheta_dpa
#             )
#     else:
#         if (p_s == airlayer.p_a).sum() == len(airlayer.p_a):
#             dH_dpa = (1 / r_H) * ( 
#                 ( ( drho_dpa * airlayer.c_p + dcp_dpa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) )
#                 + airlayer.rho_a * airlayer.c_p * dtheta_dpa
#             )
#         else:
#             dH_dpa = (1 / r_H) * ( 
#                 ( ( drho_dpa * airlayer.c_p + dcp_dpa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) )
#                 - airlayer.rho_a * airlayer.c_p * dtheta_dpa
#             )    

#     return dH_dpa

def calc_dH_dpa(airlayer, r_H, theta_s, p_s=None):

    # theta_s = T_s * (airlayer.theta_a / airlayer.T_a)

    drho_dpa = calc_drho_dpa(airlayer)
    dcp_dpa = calc_dcp_dpa(airlayer)

    dtheta_dpa = calc_dtheta_dpa(airlayer)

    dH_dpa = (1 / r_H) * ( 
        ( ( drho_dpa * airlayer.c_p + dcp_dpa * airlayer.rho_a ) * (theta_s - airlayer.theta_a) )
        - airlayer.rho_a * airlayer.c_p * dtheta_dpa
    ) 

    return dH_dpa

def calc_dH_dps(airlayer, r_H, theta_s, p_s):


    if isinstance(p_s, float):
        if p_s == airlayer.p_a:
            dH_dps = calc_dH_dpa(airlayer, r_H, theta_s, p_s)
        else:
            dH_dps = ((airlayer.rho_a * airlayer.c_p) / r_H ) * (- (R_D / c_pd) * (theta_s / p_s))
    else:
        if (p_s == airlayer.p_a).sum() == len(airlayer.p_a):
            dH_dps = calc_dH_dpa(airlayer, r_H, theta_s, p_s)
        else:
            dH_dps = ((airlayer.rho_a * airlayer.c_p) / r_H ) * (- (R_D / c_pd) * (theta_s / p_s))

    return dH_dps



#-------------------------------------------------------------------------------
# LATENT HEAT FLUX PARTIAL DERIVATIVES
#-------------------------------------------------------------------------------

# SURFACE TEMPERATURE
def calc_dLE_dTs(airlayer, r_H, epsilon_s, T_s, p_s):

    # theta_coeff = airlayer.theta_a/airlayer.T_a
    # theta_coeff = (100. / p_s) ** (R_D / c_pd)
    dtheta_dTa = calc_dtheta_dTa(airlayer)
    dLWout_dTs = calc_dLWout_dTs(T_s, epsilon_s)

    dLE_dTs = dLWout_dTs - (airlayer.rho_a * airlayer.c_p * dtheta_dTa / r_H)

    return dLE_dTs

# RESISTANCE TO HEAT TRANSFER
def calc_dLE_dra(airlayer, r_H, theta_s):

    # theta_coeff = airlayer.theta_a/airlayer.T_a
    # theta_s = theta_coeff * T_s

    dLE_drH = airlayer.rho_a * airlayer.c_p * ((theta_s - airlayer.theta_a) / (r_H **2))

    return dLE_drH

# SHORTWAVE IN
def calc_dLE_dRsw(albedo):
    
    dLE_dRsw = 1 - albedo

    return dLE_dRsw

# AIR TEMPERATURE
def calc_dLE_dTa(airlayer, r_H, theta_s):

    dLWin_dTa = calc_dLWin_dTa(airlayer)
    dH_dTa = calc_dH_dTa(airlayer, r_H, theta_s)

    dLE_dTa = dLWin_dTa - dH_dTa

    return dLE_dTa

# RELATIVE HUMIDITY
def calc_dLE_dhr(airlayer, r_H, theta_s):

    dLWin_dhr = calc_dLWin_dhr(airlayer)
    dH_dhr = calc_dH_dhr(airlayer, r_H, theta_s)

    dLE_dhr = dLWin_dhr - dH_dhr

    return dLE_dhr

# AIR PRESSURE
def calc_dLE_dpa(airlayer, r_H, theta_s, p_s):

    dLE_dpa = - calc_dH_dpa(airlayer, r_H, theta_s, p_s)

    return dLE_dpa

# SURFACE PRESSURE
def calc_dLE_dps(airlayer, r_H, theta_s, p_s):

    dLE_dps = - calc_dH_dps(airlayer, r_H, theta_s, p_s)

    return dLE_dps



# WIND SPEED
def calc_dLE_du(airlayer, r_aH, theta_s):

    # theta_s = T_s * (airlayer.theta_a / airlayer.T_a)
    dLE_du = - airlayer.rho_a * airlayer.c_p * ((theta_s - airlayer.theta_a) / (airlayer.u * r_aH))

    return dLE_du

# SOIL HEAT FLUX
def calc_dLE_dG(G=0.):
    return -1.

#-------------------------------------------------------------------------------
# BOWEN RATIO
#-------------------------------------------------------------------------------

def calc_dbeta_dTa(air1, air2, d_i, gamma_i=2):

    air_dict = {1 : air1, 2 : air2}

    c_p = air_dict.get(gamma_i).c_p 
    lambda_v = air_dict.get(gamma_i).lambda_v


    dtheta_dTa = calc_dtheta_dTa( air_dict.get(d_i) )
    dq_dTa = calc_dq_dTa( air_dict.get(d_i) )

    dbeta_dT = (
        (c_p / lambda_v) *
        ( 1 / ( (air2.q - air1.q)**2 ) ) *
        ( ( (air2.q - air1.q) * dtheta_dTa ) - ( (air2.theta_a - air1.theta_a) * dq_dTa ) )
    )

    if d_i == 1:
        dbeta_dT = dbeta_dT * -1

    if d_i == gamma_i:

        dcp_dTa = calc_dcp_dTa( air_dict.get(d_i) )
        dlv_dTa = calc_dlhvap_dTa( air_dict.get(d_i) )
        
        ex = ( 
            ( (air2.theta_a - air1.theta_a) / (air2.q - air1.q) ) *
            ( lambda_v * dcp_dTa - c_p * dlv_dTa) *
            ( 1 / (lambda_v**2) ) 
        )
    
        dbeta_dT = dbeta_dT + ex

    return dbeta_dT

def calc_dbeta_dhr(air1, air2, d_i, gamma_i=2):
    air_dict = {1 : air1, 2 : air2}
    
    c_p = air_dict.get(gamma_i).c_p 
    lambda_v = air_dict.get(gamma_i).lambda_v

    dq_dhr = calc_dq_dhr( air_dict.get(d_i) )

    dbeta_dhr = (
        (c_p / lambda_v) *
        ( 1 / ( (air2.q - air1.q)**2 ) ) *
        ( (air2.theta_a - air1.theta_a) * dq_dhr )
    )

    if d_i == 2:
        dbeta_dhr = dbeta_dhr * -1
    
    if d_i == gamma_i:
        dcp_dhr = calc_dcp_dhr(air_dict.get(d_i))
        
        ex = ( 
            ( (air2.theta_a - air1.theta_a) / (air2.q - air1.q) ) *
            ( (1 / lambda_v) * dcp_dhr)
        )
    
        dbeta_dhr = dbeta_dhr + ex
    
    return dbeta_dhr

def calc_dbeta_dpa(air1, air2, d_i, gamma_i=2):
    air_dict = {1 : air1, 2 : air2}
    
    c_p = air_dict.get(gamma_i).c_p 
    lambda_v = air_dict.get(gamma_i).lambda_v

    dtheta_dpa = calc_dtheta_dpa( air_dict.get(d_i) )
    dq_dpa = calc_dq_dpa( air_dict.get(d_i) )

    dbeta_dpa = (
        (c_p / lambda_v) *
        ( 1 / ( (air2.q - air1.q)**2 ) ) *
        ( ((air2.q - air1.q) * dtheta_dpa) - ((air2.theta_a - air1.theta_a) * dq_dpa) )
    )

    if d_i == 1:
        dbeta_dpa = dbeta_dpa * -1

    if d_i == gamma_i:

        dcp_dpa = calc_dcp_dpa( air_dict.get(d_i) )
        
        ex = ( 
            ( (air2.theta_a - air1.theta_a) / (air2.q - air1.q) ) *
            ( (1 / lambda_v) * dcp_dpa)
        )
    
        dbeta_dpa = dbeta_dpa + ex

    return dbeta_dpa

def calc_dbeta_dG(G=0.):
    return 0.

#-------------------------------------------------------------------------------
# AVAILABLE ENERGY
#-------------------------------------------------------------------------------

def calc_dQav_dTs(T_s, epsilon_s=0.98):

    dQav_dTs = (-4 * SIGMA * epsilon_s * (T_s**3))

    return dQav_dTs

def calc_dQav_dTa(airlayer):

    dQav_dTa = calc_dLWin_dTa(airlayer)

    return dQav_dTa

def calc_dQav_dhr(airlayer):

    dQav_dhr = calc_dLWin_dhr(airlayer)

    return dQav_dhr

def calc_dQav_dG(G=0.):
    return -1.

#-------------------------------------------------------------------------------
# LE FROM BOWEN RATIO
#-------------------------------------------------------------------------------

def calc_dLEbr_dQav(beta):

    dLEbr_dQav = 1 / (1 + beta)

    return dLEbr_dQav

def calc_dLEbr_dbeta(beta, Q_av):
    
    dLEbr_dbeta = - ( Q_av) / ( (1 + beta)**2 )

    return dLEbr_dbeta

dbeta_dict = {
    'T_a' : calc_dbeta_dTa,
    'h_r' : calc_dbeta_dhr,
    'p_a' : calc_dbeta_dpa,
    'G' : calc_dbeta_dG
}

dQav_dict = {
    'T_s' : calc_dQav_dTs,
    'T_a' : calc_dQav_dTa,
    'h_r' : calc_dQav_dhr,
    'G' : calc_dQav_dG
}

def calc_dLEbr_dx(var, beta, Q_av, air1, air2, d_i, gamma_i):

    if d_i == 1:
        dQav_dx = dQav_dict.get(var, lambda _ : 0)(air1)
    else:
        dQav_dx = 0

    dbeta_dx = dbeta_dict.get(var)(air1, air2, d_i, gamma_i)

    dLEbr_dx = ( ((1 + beta) * dQav_dx) - (Q_av * dbeta_dx) ) / ( (1 + beta)**2 )

    return dLEbr_dx

def calc_dLEbr_dTs(beta, T_s, epsilon_s=0.98):

    dLEbr_dTs = calc_dQav_dTs(T_s, epsilon_s) / (1 + beta)

    return dLEbr_dTs

def calc_dLEbr_dRsw(albedo, beta):

    dLEbr_dRsw = (1 - albedo) / (1 + beta)
    
    return dLEbr_dRsw



#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'MIT'
__date__ = 'Wed 24 Aug 22 13:39:37'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           model.py
Compatibility:  Python 3.7.0
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2022


"""


# IMPORTS
import numpy as np
import pandas as pd
import xarray as xr

from . import utils

# VARIABLES
KAPPA = 0.41        # von Karman constant [unitless]
G = 9.8             # acceleration due to gravity [m s-2]

flag_dict = {
    'NO_FLAG' : (0, 'All fluxes are good.'),
    'F_NEG_LE' : (1, 'Negative latent heat flux; forced to zero'),
    'F_NEG_H' : (2, 'Negative sensible heat flux; forced to zero.'),
    'F_ERROR' : (255, 'Arithmetic error; bad data; values should be discarded.')
}


class FluxModel:
    """
    _summary_
    """

    def __init__(self, airlayer, surface, radiation):
        """
        Parameters
        ----------
        airlayer : AirLayer
            air.AirLayer object for the model. Contains the following variables:
            z, u, T_a, p_a, e_a, c_p, h_r, 
        surface : Surface
            surface.Surface object for the model. Contains the following variables:
            h, d0, z_0m, z_0h, T_s
        radiation : Radiation
            radiation.Radiation object for the model. Contains the following variables:
            R_n, G, [SW_IN, SW_OUT, LW_IN, LW_OUT]
        """

        self.air = airlayer             # z, u, T_a, p_a, e_a, c_p, h_r, 
        self.surface = surface          # h, d0, z_0m, z_0h
        self.radiation = radiation      # R_n, G
        # self.theta_s = self.air.calc_theta(self.surface.T_s)
        # self._set_theta_s()

    def __repr__(self):
        class_name = type(self).__name__
        return '{}'.format(class_name)
    
    # def _set_theta_s(self):
    #     """
    #     Set potential surface temperature (self.surface.theta_s) for the model. 
    #     NOTE: Probably should go somewhere else?
    #     """
    #     # NOTE: It's probably worth forcing these to the same resolution regardless, 
    #     # but for now, only doing it when necessary bc resampling takes awhile.
    #     if self.radiation.R_n.shape != self.surface.T_s.shape:
    #         self.theta_s = self.air.calc_theta(
    #             self.surface.T_s.interp(x=self.radiation.R_n.x, y=self.radiation.R_n.y)
    #         )
    #     else:
    #         self.theta_s = self.air.calc_theta(self.surface.T_s)


    def get_var_stats(self, var):
        """
        Get summary statistics (min, max, mean, median, standard deviation) for 
        a variable.

        Parameters
        ----------
        var : xarray.DataArray | xarray.DataSet
            Pixel-wise variable for which to calculate summary statistics

        Returns
        -------
        stats_dict : dict
            Summary statistics for var as a dictionary.

        # NOTE: Probably will refactor.
        """
        
        stats_dict = {
            'MAX' : var.max().item(),
            'MIN' : var.min().item(),
            'MEAN' : var.mean().item(),
            'MEDIAN' : var.median().item(),
            'STD' : var.std().item()
        }

        return stats_dict

    def get_summary_stats(self, as_df=True, mask=None):
        """
        Get summary statistics for all flux-related variables (zeta, u_star, 
        Psi_M, Psi_H, r_aH, L, H, LE). Note that only array-like variables will
        return summary statistics; others will return the scalar values.

        Returns
        -------
        all_dict : dict
            Dictionary of dictionaries containing flux variables (keys) and their
            summary statistics (values : dict).
        
        # TODO: Change the way non-arrays are handled (this is sloppy--all should be the same type).
        # TODO: Refactor--this is temporary and will likely want to change.
        """
        # Get flux variables as dict
        var_dict = self.get_fluxes()
        # Initialize empty dictionary for variables
        all_dict = {}
        # Iterate through flux variables and return summary stats
        for key,val in var_dict.items():
            try:
                if mask is not None:
                    all_dict[key] = self.get_var_stats(val.where(mask))
                else:
                    all_dict[key] = self.get_var_stats(val)
            except:
                all_dict[key] = {'MEAN' : val}
        
        if as_df:
            return self.get_stats_df(all_dict)
        else:
            return all_dict

    def get_stats_df(self, stats_dict):

        # stats = self.get_summary_stats()

        df = pd.DataFrame(stats_dict).T.reset_index().rename(columns={'index':'Variable'})

        return df

    def get_fluxes(self):
        """
        Get flux-related variables created/output by the model.

        Returns
        -------
        dict
            Dictionary of flux variables calculated by the model run.
        """
        flux_dict =  {
            'zeta' : self.zeta,
            'u_star' : self.u_star,
            'Psi_M' : self.Psi_M,
            'Psi_H' : self.Psi_H,
            'r_aH' : self.r_aH,
            'r_bH' : self.r_bH,
            'r_H' : self.r_H,
            'L' : self.L,
            'H' : self.H,
            'LE' : self.LE,
            'ET' : self.calc_ET(self.LE),
            'EF' : self.calc_EF(self.LE)
            # 'n_iter' : len(self.Ls)
        }
        # TODO: Fix below
        for arr in flux_dict.values():
            try:
                utils.set_crs(arr, self.surface.theta_s)
            except:
                pass

        return flux_dict
    
    # def set_crs(self, arr):

    #     arr.rio.set_crs(self.surface.theta_s.rio.crs)

    def calc_ET(self, LE):

        self.ET = (LE / self.air.lambda_v) * 3600

        return self.ET
    
    def calc_EF(self, LE):
            
        self.EF = (LE / (self.radiation.R_n - self.radiation.G) )
        

        return self.EF

    def _run(self, L=np.inf, use_psi_0=True, allow_neg_flux=True):
        """
        Run model (defaults to neutral).

        Parameters
        ----------
        L : float, optional
            Initial Monin-Obukhov length [m]. The default is np.inf.
        use_psi_0 : bool, optional
            Whether to use the Psi_0 values (makes a minimal difference in r_aH).
            The default is False.
        allow_neg_flux : bool, optional
            Whether or not to allow negative values for H and LE. If False, negative
            values of both H and LE will be set to 0. The default is True (for now).
        
        Yields
        ------
        self.zeta : float
            Characteristic length scale.
        self.Psi_M : float
            Stability correction term for momentum transport.
        self.Psi_H : float
            Stability correction term for heat/vapor transport.
        self.u_star : float
            Friction velocity [m s-1]
        self.r_aH : float
            Aerodynamic resistance to heat/vapor transport.
        self.H : array-like
            Sensible heat flux [W m-2].
        self.LE : array-like
            Latent heat flux [W m-2].
        self.L : array-like
            Monin-Obukhov length [m].
        """
        # 1. Calculate zeta
        self.zeta = self.calc_zeta(L)
        # 2. Calculate Psi terms
        self.Psi_M = self.calc_Psi_M(self.zeta)
        self.Psi_H = self.calc_Psi_H(self.zeta)
        # Optionally calculate Psi_0 terms
        if use_psi_0:
            Psi_M0 = self.calc_Psi_M(self.surface.z_0m / L)
            Psi_H0 = self.calc_Psi_H(self.surface.z_0h / L)
        else:
            Psi_M0 = 0
            Psi_H0 = 0
        # 3. Calculate u_star
        self.u_star = self.calc_ustar(self.Psi_M, Psi_m0=Psi_M0)
        # 4. Calculate r_aH
        self.r_aH = self.calc_r_aH(self.Psi_M, self.Psi_H, Psi_M0=Psi_M0, Psi_H0=Psi_H0)
        # 5. Calculate r_bH
        self.r_bH = self.calc_r_bH(u_star=self.u_star, L=L, c=90.)
        # 6. Calculate total resistance, r_H
        self.r_H = self.calc_r_H(r_bH=self.r_bH)
        # 5. Calculate H
        self.H = self.calc_H(self.surface.theta_s, self.r_H)
        # Flag bad data
        if not allow_neg_flux:
            self.flag = np.where(self.H < 0, flag_dict.get('F_NEG_H')[0], self.flag)
            self.H = xr.where(self.H < 0.0, 0.0, self.H)
        # 6. Calculate LE
        self.LE = self.calc_LE(self.H)
        # Flag bad data
        # if not allow_neg_flux:
        #     self.flag = np.where(self.LE < 0, flag_dict.get('F_NEG_LE')[0], self.flag)
        #     self.LE = xr.where(self.LE < 0.0, 0.0, self.LE)
        self.flag = np.where(self.LE < 0, flag_dict.get('F_NEG_LE')[0], self.flag)
        self.LE = xr.where(self.LE < 0.0, 0.0, self.LE)
        # 7. Recalculate L
        H_v = self.calc_Hv(self.H, self.LE)
        # self.L = self.calc_L(self.u_star, H_v.where(H_v != 0.0, other=np.nan)) # change to calc_L on mean of H_v
        self.L = self.calc_L(self.u_star, np.nanmean(H_v))
        # Calculate difference in L
        #  L_diff = abs(np.nanmean(self.L.where(abs(self.L) < 1e6, other=np.nan)) - L)
        L_diff = abs(self.L - L)

        return L_diff



        
    # def run(self, L=np.inf, use_psi_0=True, allow_neg_flux=False, get_fluxes=False):
    def run(self, L=np.inf, use_psi_0=True, allow_neg_flux=False, max_i=15, L_thresh=0.001, get_fluxes=False):
        """
        Actually run the model and optionally return output variables.

        Parameters
        ----------
        L : float, optional
            Initial Monin-Obukhov length [m]. The default is np.inf.
        use_psi_0 : bool, optional
            Whether to use the Psi_0 values (makes a minimal difference in r_aH).
            The default is True.
        allow_neg_flux : bool, optional
            Whether or not to allow negative values for H and LE. If False, negative
            values of both H and LE will be set to 0. The default is True (for now).
        get_fluxes : bool, optional
            Whether or not to return flux variables as a dictionary. The default is False.

        Yields
        ------
        self.zeta : float
            Characteristic length scale.
        self.Psi_M : float
            Stability correction term for momentum transport.
        self.Psi_H : float
            Stability correction term for heat/vapor transport.
        self.u_star : float
            Friction velocity [m s-1]
        self.r_aH : float
            Aerodynamic resistance to heat/vapor transport.
        self.H : array-like
            Sensible heat flux [W m-2].
        self.LE : array-like
            Latent heat flux [W m-2].
        self.L : array-like
            Monin-Obukhov length [m].

        """
        self.flag = flag_dict.get('NO_FLAG')[0]
        # self._run(L=L, use_psi_0=use_psi_0, allow_neg_flux=allow_neg_flux)
        
        L_diff = np.inf
        self.L = L

        for i in range(max_i):
            L_diff = self._run(L=np.nanmean(self.L), use_psi_0=use_psi_0, allow_neg_flux=allow_neg_flux)
            if L_diff <= L_thresh:
                print(f"Finished {i+1} iterations with an L diff: {L_diff}")
                break
            elif i == 1:
                print(f"H : {np.nanmean(self.H)}, LE : {np.nanmean(self.LE)}")
            else:
                print(f"L diff (n = {i+1}): {L_diff}")
        self._n_iter = i + 1

        if get_fluxes:
            return self.get_fluxes()    

    


    def _run_pix(self, L=np.inf, use_psi_0=True, mask=None, allow_neg_flux=False):

        if mask is not None:
            self.zeta = xr.where(mask, self.calc_zeta(L), self.zeta)
        else:
            self.zeta = self.calc_zeta(L)

        self.Psi_M = self.calc_Psi_M(self.zeta)
        self.Psi_H = self.calc_Psi_H(self.zeta)

        if use_psi_0:
            Psi_M0 = self.calc_Psi_M(self.surface.z_0m / L)
            Psi_H0 = self.calc_Psi_H(self.surface.z_0h / L)
        else:
            Psi_M0 = 0
            Psi_H0 = 0

        # 3. Calculate u_star
        self.u_star = self.calc_ustar(self.Psi_M, Psi_m0=Psi_M0)
        # 4. Calculate r_aH
        self.r_aH = self.calc_r_aH(self.Psi_M, self.Psi_H, Psi_M0=Psi_M0, Psi_H0=Psi_H0)
        # 5. Calculate r_bH
        self.r_bH = self.calc_r_bH(u_star=self.u_star, L=self.L, c=90.)
        # 6. Calculate total resistance, r_H
        self.r_H = self.calc_r_H(r_bH=self.r_bH)
        # 5. Calculate H
        self.H = self.calc_H(self.surface.theta_s, self.r_aH)
        # Flag bad data
        if not allow_neg_flux:
            self.flag = np.where(self.H < 0, flag_dict.get('F_NEG_H')[0], self.flag)
            self.H = xr.where(self.H < 0.0, 0.0, self.H)
        # 6. Calculate LE
        self.LE = self.calc_LE(self.H)
        # Flag bad data
        if not allow_neg_flux:
            self.flag = np.where(self.LE < 0, flag_dict.get('F_NEG_LE')[0], self.flag)
            self.LE = xr.where(self.LE < 0.0, 0.0, self.LE)
        # 7. Recalculate L
        H_v = self.calc_Hv(self.H, self.LE)
        self.L = self.calc_L(self.u_star, H_v.where(H_v != 0.0))
        # Calculate difference in L
        L_diff = abs(self.L - L)

        return L_diff

    def run_pix(self, L=np.inf, use_psi_0=True, max_i=15, L_thresh=0.001, allow_neg_flux=False):

        L_diff = np.inf
        self.L = L
        self.zeta = self.calc_zeta(L)

        self.flag = flag_dict.get('NO_FLAG')[0]

        for i in range(max_i):
            # Create mask where L_diff > threshold = True, else False (i.e. don't re-calculate)
            mask = np.logical_and(L_diff >= L_thresh, self.flag != flag_dict.get('F_ERROR')[0])
            n_pix = mask.sum().item()
            print(f"Iteration {i}, non-converged pixels: {n_pix} . max L_diff : {np.nanmax(L_diff)}")
            # Run model again and get difference in L
            L_diff = self._run_pix(L=self.L, use_psi_0=use_psi_0, mask=mask, allow_neg_flux=allow_neg_flux)
            # If all values are lower than the threshold, stop iterating.
            if np.nanmax(L_diff) <= L_thresh:
                print(f"Finished iteration with a max L diff: {np.nanmax(L_diff)}")
                break
        # Set values outside image to NaN in zeta (they are 0 because of the first iteration)
        self.zeta = xr.where(self.surface.theta_s.notnull(), self.zeta, np.nan)
        self._n_iter = i + 1




    def calc_zeta(self, L : float):
    # def calc_zeta(z, d_0, L):
        """
        Calculate the characteristic length scale (zeta).

        Parameters
        ----------
        L : float
            Monin-Obukhov length [m].

        Returns
        -------
        zeta : float
            Characteristic length scale.
        """

        zeta = (self.air.z - self.surface.d_0) / L

        # try:
        #     zeta = zeta.item()
        # except:
        #     pass

        return zeta
    
    def calc_Psi_M(self, zeta):
        raise NotImplementedError
    
    def calc_Psi_H(self, zeta):
        raise NotImplementedError
    
    def calc_ustar(self, Psi_m, Psi_m0=0):
    # def calc_ustar(u, z, d_0, z_0m, Psi_m, Psi_m0=0):
        """
        Calculate friction velocity [m s-1].

        Parameters
        ----------
        Psi_m : float
            Stability correctoin function for momentum transport
        Psi_m0 : float, optional
            Additional stability correction term for momentum. The default is 0.
        
        Additional Variables
        --------------------
        u : float
            Wind speed [m s-1]
        z : float
            Height of wind speed measurement [m].
        h : float
            Canopy height [m].

        Returns
        -------
        u_star : float
            Friction velocity [m s-1].
        """

        u_star = (self.air.u * KAPPA) / ( np.log( (self.air.z - self.surface.d_0) / self.surface.z_0m ) - Psi_m + Psi_m0)

        return u_star

    def calc_r_aH(self, Psi_M, Psi_H, Psi_M0=0, Psi_H0=0):
    # def calc_r_a(u, z, d_0, z_0m, z_0h, Psi_M, Psi_H):
        """
        Calculate aerodynamic resistance according to MOST.

        Parameters
        ----------
        Psi_M : float
            Stability correction factor for momentum transport.
        Psi_H : float
            Stability correction factor for heat transport.
        Psi_M0 : float, optional
            Additional stability correction term for momentum. The default is 0.
        Psi_H0 : float, optional
            Additional stability correction term for heat. The default is 0.

        Additional Variables
        --------------------
        u : float
            Wind speed [m s-1].
        z : float
            Height of wind speed measurement.
        d_0 : float
            Zero-plane displacement height [m].
        z_0m : float
            Roughness length for momentum transport [m].
        z_0h : float
            Roughness length for heat transport [m].

        Returns
        -------
        r_a : float
            Aerodynamic resistance to heat transport [s m-1].
        """

        # Calculate aerodynamic resistance
        r_aH = ( 
            ( (np.log((self.air.z - self.surface.d_0) / self.surface.z_0m) - Psi_M + Psi_M0) * 
            (np.log((self.air.z - self.surface.d_0) / self.surface.z_0h) - Psi_H + Psi_H0) )
            / (KAPPA**2 * self.air.u) 
        )

        return r_aH


    def calc_r_bH(self, u_star, L, c=90.):
        """
        Calculate boundary layer resistance.
        """

        u_h = self.calc_u_z(z=self.surface.h, u_star=u_star, L=L)
        u_0m = self.calc_u0m(u_h = u_h)

        r_bH = (c / self.surface.lai) * ( (self.surface.w_l / u_0m) ** (1/2) )

        # TODO: Figure out a better way to do this. For now, just using a low-ish 
        # threshold of NDVI to avoid calculating r_bH where LAI is ~ 0 (which makes
        # r_bH –> infinity).
        # r_bH = xr.where(self.surface.ndvi < 0.25, np.nan, r_bH)
        r_bH = xr.where(self.surface.lai < 1., np.nan, r_bH)


        return r_bH

    def calc_u_z(self, z, u_star, L):

        Psi_m2 = self.calc_Psi_M( (z - self.surface.d_0) / L )
        Psi_m1 = self.calc_Psi_M( (self.air.z - self.surface.d_0) / L )

        u_z = ( ( 
            (u_star / KAPPA) * 
            ( np.log((z - self.surface.d_0) / (self.air.z - self.surface.d_0) ) 
             -  Psi_m2 + Psi_m1 ) 
            ) + self.air.u )

        return u_z

    def calc_u0m(self, u_h):

        a = ( 0.28 * 
             (self.surface.lai**(2/3)) * (self.surface.h**(1/3)) * 
             (self.surface.w_l**(-1/3)) 
        )
        
        u_0m = u_h * np.exp(
            -a * (1 - ( (self.surface.d_0 + self.surface.z_0m) / self.surface.h ) ) 
        )

        return u_0m


    def calc_r_H(self, r_bH=0.):
        r_H = self.r_aH + r_bH
        return r_H


    def calc_H(self, theta_s, r_aH):
    # def calc_H(self, rho_a, c_p, theta_s, theta_a, r_aH):
        """
        Calculate sensible heat flux using a temperature gradient and aerodynamic 
        resistance to heat transport.

        Parameters
        ----------
        theta_s : float or array-like
            Potential surface temperature [K]
        r_aH : float
            Aerodynamic resistance to heat transport [s m-1]

        Additional Variables
        --------------------
        rho_a : float
            Air density [kg m-3]
        c_p : float
            Heat capacity of air [J K-1 kg-1]
        theta_a : float
            Potential air temperature [K]

        Returns
        -------
        H : float or array-like
            Sensible heat flux [W m-2]
        """
        H = self.air.rho_a * self.air.c_p * ( (theta_s - self.air.theta_a) / r_aH )

        return H
    
    def calc_LE(self, H):
    # def calc_LE(R_n, G, H):
        """
        Calculate latent heat flux as the residual of the surface energy balance equation.

        Parameters
        ----------
        H : float or array
            Sensible heat flux [W m-2].
        
        Additional Variables
        --------------------
        R_n : float or array
            Net radiation [W m-2].
        G : float or array
            Soil heat flux [W m-2].

        Returns
        -------
        LE : float or array
            Latent heat flux [W m-2].
        """

        LE = xr.where(H < 0, self.radiation.R_n - self.radiation.G, self.radiation.R_n - self.radiation.G - H)
        
        # LE = self.radiation.R_n - self.radiation.G - H
        
        if LE.shape != H.shape:
            LE = LE.interp(x=H.x, y=H.y)

        return LE
    
    def calc_L(self, u_star, H_v):
    # def calc_L(self, u_star, theta_v, rho_a, c_p, H_v):
        """
        Calculate the Monin-Obukhov length, L [m].

        Parameters
        ----------
        u_star : float
            Friction velocity, by default None
        H_v : float
            Virtual sensible heat flux [W m-2], by default None
        
        Additional Variables
        --------------------
        theta_v : float
            Virtual potential temperature [K], by default None
        rho_a : float
            Air density [kg m-3], by default None
        c_p : float
            Heat capacity of air at constant pressure [J kg-1 K-1]

        Returns
        -------
        L : float
            Monin-Obukhov length [m].
        """

        # theta_v = met.calc_theta(T_a, p_a)

        L = - ( ( ( u_star**3 ) * self.air.theta_v ) / 
               ( KAPPA * G * ( H_v / ( self.air.rho_a * self.air.c_p ) ) ) )

        return L

    def calc_Hv(self, H, LE):
    # def calc_Hv(H, LE, T_a, p_a, e_a):
        """
        Calculate virtual sensible heat flux.

        Parameters
        ----------
        H : float | array-like
            Sensible heat flux [W m-2]
        LE : float | array-like
            Latent heat flux [W m-2]
        
        Additional Variables
        --------------------
        T_a : float
            Temperature [K]
        c_p : float
            Heat capacity of air at constant pressure [J kg-1 K-1]
        theta_a : float
            Potential temperature [K]

        Returns
        -------
        H_v : float
            Virtual sensible heat flux [W m-2]
        """
        # theta = met.calc_theta(T_a, p_a)
        E = LE / self.air.lambda_v
        # c_p = met.calc_cp(p_a, e_a)

        H_v = H + ( 0.61 * self.air.c_p * self.air.theta_a * E )

        return H_v




class NeutralModel(FluxModel):
    """
    Calculate turbulent fluxes assuming neutral conditions.
    """

    def __init__(self, airlayer, surface, radiation):

        super().__init__(airlayer, surface, radiation)

        self.L = np.inf

    # def run(self, n_iter=1):
    #     self._run(n_iter=n_iter)
    def run(self, L=np.inf, use_psi_0=False, allow_neg_flux=False, get_fluxes=False):
        self.flag = flag_dict.get('NO_FLAG')[0]
        self._run(L=L, use_psi_0=use_psi_0, allow_neg_flux=allow_neg_flux)
        if get_fluxes:
            return self.get_fluxes()    


    def run_pix(self, L=np.inf, use_psi_0=False, max_i=1, L_thresh=0.001, allow_neg_flux=False):
        return super().run_pix(L, use_psi_0, max_i, L_thresh, allow_neg_flux)

    def calc_Psi_M(self, zeta):
        return 0
    
    def calc_Psi_H(self, zeta):
        return 0




class DyerModel(FluxModel):
    """
    Calculate turbulent fluxes using the equations for stability correction terms
    found in Bonan (2016). NOTE: Not original source.
    """

    def __init__(self, airlayer, surface, radiation):

        super().__init__(airlayer, surface, radiation)
    
    # def run(self, L=np.inf):
    #     # Run first time to get neutral L
    #     self._run(L=L)
    #     # Run again to get actual fluxes
    #     self._run(L=self.L.mean())


    def calc_Psi_M(self, zeta):
        """
        Calculates the stability correction function for momentum transport.

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        psi_m : float
            Stability correction function for momentum transport.

        Reference: ???? Bonan (but not original source)
        """

        Psi_m = xr.where(zeta < 0, self._calc_Psi_M_unstable(zeta), self._calc_Psi_stable(zeta))

        return Psi_m

    def calc_Psi_H(self, zeta):
        """
        Calculates the stability correction function for heat (and vapour) transport.

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        psi_h : float
            Stability correction function for heat (or vapor) transport.

        Reference: ??? Bonan (but not original source)
        """

        Psi_h = xr.where(zeta < 0, self._calc_Psi_H_unstable(zeta), self._calc_Psi_stable(zeta))

        return Psi_h

    def _calc_Psi_stable(self, zeta):
        """
        Calculate Psi for stable conditions (zeta >= 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        psi : float
            Stability correction term for momentum and heat (or vapor) transport
            under stable conditions.
        """

        Psi_m = -5 * zeta

        return Psi_m
    
    def _calc_Psi_M_unstable(self, zeta):
        """
        Calculate Psi_M for unstable conditions (zeta < 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        Psi_m : float
            Stability correction term for momentum transport under unstable conditions.
        """
        x = (1 - 16 * zeta)**(1/4)

        Psi_m = 2 * np.log((1 + x) / 2) + np.log( (1 + x**2) / 2) - 2 * np.arctan(x) + np.pi/2

        return Psi_m
    
    def _calc_Psi_H_unstable(self, zeta):
        """
        Calculate Psi_H for unstable conditions (zeta < 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        Psi_m : float
            Stability correction term for heat transport under unstable conditions.
        """
        x = (1 - 16*zeta)**(1/4)
        Psi_h = 2 * np.log((1 + x**2)/2)
        
        return Psi_h



class BrutsaertModel(FluxModel):
    """
    Calculate turbulent fluxes using the stability correction terms found in 
    Brutsaert (2005). 
    """
    def __init__(self, airlayer, surface, radiation):

        super().__init__(airlayer, surface, radiation)

    # def run(self, L=np.inf):
    #     # Run first time to get neutral L
    #     self._run(L=L)
    #     # Run again to get actual fluxes
    #     self._run(L=self.L.mean())


    def calc_Psi_M(self, zeta):
        """
        Calculates the stability correction function for momentum transport.

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        psi_m : float
            Stability correction function for momentum transport.

        Reference: Brutsaert, W. (2005). Hydrology: An Introduction. Cambridge University Press.
        """
        # # Stable conditions, zeta >= 0
        # if zeta >= 0:
        #     a = 6.1
        #     b = 2.5
        #     Psi_m = -a * np.log(zeta + (1.0 + zeta**b)**(1.0/b))    # Eq. (2.59)

        # # Unstable conditions, zeta < 0
        # else:
        #     a = 0.33
        #     b = 0.41
        #     y = -1 * zeta
        #     x = (y / a)**(1/3)

        #     Psi_0 = -np.log(a) + np.sqrt(3) * b * a**(1/3) * np.pi/6.0
        #     y = min(y, b**(-1/3))

        #     Psi_m = np.log(a+y) - 3 * b * y**(1/3) + \
        #                 ((b*a**(1/3)) / 2 ) * np.log( (1+x)**2 / (1 - x + x**2)) + \
        #                 np.sqrt(3) * b * a**(1/3) * np.arctan((2*x - 1) / np.sqrt(3) ) + \
        #                 Psi_0                       # Eq. (2.63)

        Psi_m = xr.where(zeta < 0, self._calc_Psi_M_unstable(zeta), self._calc_Psi_stable(zeta))

        # if Psi_m.ndim == 1:
        #     return Psi_m
        # else:
        #     return Psi_m.item()
        return Psi_m
    
    def calc_Psi_H(self, zeta):
        """
        Calculates the stability correction function for heat (and vapour) transport.

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        psi_h : float
            Stability correction function for heat (or vapor) transport.

        Reference: Brutsaert, W. (2005). Hydrology: An Introduction. Cambridge University Press.
        """
        # # Stable conditions, zeta >=0
        # if zeta >= 0:
        #     Psi_h = self.calc_Psi_M(zeta)

        # # Unstable conditions, zeta < 0
        # else:
        #     c = 0.33
        #     d = 0.057
        #     n = 0.78
        #     y = -1 * zeta

        #     Psi_h = ((1 - d) / n) * np.log((c + y**n) / c )     # Eq. (2.64)

        Psi_h = xr.where(zeta < 0, self._calc_Psi_H_unstable(zeta), self._calc_Psi_stable(zeta))

        # if Psi_h.ndim == 1:
        #     return Psi_h
        # else:
        #     return Psi_h.item()
        return Psi_h
    
    def _calc_Psi_stable(self, zeta):
        """
        Calculate Psi for stable conditions (zeta >= 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        Psi : float
            Stability correction term for momentum or heat transport under stable 
            conditions.
        """
        a = 6.1
        b = 2.5
        Psi_m = -a * np.log(zeta + (1.0 + zeta**b)**(1.0/b))    # Eq. (2.59)

        return Psi_m
    
    def _calc_Psi_M_unstable(self, zeta):
        """
        Calculate Psi_M for unstable conditions (zeta < 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        Psi_m : float
            Stability correction term for momentum transport under unstable conditions.
        """
        a = 0.33
        b = 0.41
        y = -1 * zeta
        x = (y / a)**(1/3)

        Psi_0 = -np.log(a) + np.sqrt(3) * b * a**(1/3) * np.pi/6.0
        # y = min(y, b**(-1/3))
        y = np.where(y < b**(-1/3), y, b**(-1/3))

        Psi_m = np.log(a+y) - 3 * b * y**(1/3) + \
                    ((b*a**(1/3)) / 2 ) * np.log( (1+x)**2 / (1 - x + x**2)) + \
                    np.sqrt(3) * b * a**(1/3) * np.arctan((2*x - 1) / np.sqrt(3) ) + \
                    Psi_0                       # Eq. (2.63)

        return Psi_m
    
    def _calc_Psi_H_unstable(self, zeta):
        """
        Calculate Psi_H for unstable conditions (zeta < 0).

        Parameters
        ----------
        zeta : float
            Dimensionless length parameter

        Returns
        -------
        Psi_h : float
            Stability correction term for heat transport under unstable conditions.
        """
        c = 0.33
        d = 0.057
        n = 0.78
        y = -1 * zeta

        Psi_h = ((1 - d) / n) * np.log((c + y**n) / c )     # Eq. (2.64)

        return Psi_h        


class BowenModel(FluxModel):
    """
    Calculate turbulent fluxes according to the Bowen Ratio using a profile approach.
    """

    def __init__(self, airlayer_2, airlayer_1, surface, radiation):
        """
        Parameters
        ----------
        airlayer2 : AirLayer
            air.AirLayer object at height z2. Contains the following variables:
            z, u, T_a, p_a, e_a, c_p, h_r, 
        airlayer1 : AirLayer
            air.AirLayer object at height z1. Contains the following variables:
            z, u, T_a, p_a, e_a, c_p, h_r, 
        surface : Surface
            surface.Surface object for the model. Contains the following variables:
            h, d0, z_0m, z_0h, T_s
        radiation : Radiation
            radiation.Radiation object for the model. Contains the following variables:
            R_n, G, [SW_IN, SW_OUT, LW_IN, LW_OUT]
        """
        super().__init__(airlayer_2, surface, radiation)

        self.air2 = airlayer_2
        self.air1 = airlayer_1


    def run(self, allow_neg_flux=True):
        """
        Calculate turbulent fluxes according to the Bowen Ratio.

        Yields
        ------
        self.LE_hat : float
            Landscape-level average latent heat flux [W m-2]
        self.H_hat : float
            Landscape-level average sensible heat flux [W m-2]
        self.r_aH : float
            Aerodynamic resistance to heat/vapor transport [s m-2]
        self.H : array-like
            Sensible heat flux [W m-2].
        self.LE : array-like
            Latent heat flux [W m-2].
        """
        # Set flag
        self.flag = flag_dict.get('NO_FLAG')[0]
        # 1. Calculate Bowen Ratio from profile data
        beta = self.calc_beta()
        # 2. Calculate landscape-level LE from Bowen Ratio
        # self.LE_hat = self.calc_LE_bowen(beta)
        self.LE = self.calc_LE_bowen(beta)
        # 3. Calculate landscape-level H from Bowen Ratio
        # self.H_hat = self.calc_H_bowen(self.LE_hat, beta)
        self.H = self.calc_H_bowen(self.LE, beta)
        # 4. Calculate aerodynamic resistance
        self.r_H = self.calc_r_H(self.H)
        # 5. Calculate pixel-wise H
        # self.H = self.calc_H(self.surface.theta_s, self.r_H)
        # Flag bad data
        if not allow_neg_flux:
            self.flag = np.where(self.H < 0, flag_dict.get('F_NEG_H')[0], self.flag)
            self.H = xr.where(self.H < 0.0, 0.0, self.H)
        # 6. Calculate pixel-wise LE
        self.LE = self.calc_LE(self.H)
        # Flag bad data
        if not allow_neg_flux:
            self.flag = np.where(self.LE < 0, flag_dict.get('F_NEG_LE')[0], self.flag)
            self.LE = xr.where(self.LE < 0.0, 0.0, self.LE)
        # 7. Calculate L
        # H_v = self.calc_Hv(self.H, self.LE)
        # self.L = self.calc_L(self.u_star, H_v)
        # # 8. Calculate zeta
        # self.zeta = self.calc_zeta(self.L)
        self.zeta = None
        self.u_star = None
        self.Psi_M = None
        self.Psi_H = None
        self.L = None
        self.r_aH = None
        self.r_bH = None


    def calc_beta(self):
    # def calc_bowen(T_2, p_2, rh_2, T_1, p_1, rh_1):
        """
        Calculates the Bowen Ratio given the temperature and humidity of air at
        height z2 and the temperature and humidity of air at height z1.

        Parameters
        ----------
        T_2 : float
            Air temperature at height z_2 [K].
    
        p_2 : float
            Air pressure at height z_2 [kPa].

        rh_2 : float
            Relative humidity at height z_2 [%].

        T_1 : float
            Air temperature at height z_1 [K].

        p_1 : float
            Air pressure at height z_1 [kPa].
        
        rh_1 : float
            Relative humidity at height z_1 [%].

        Returns
        -------
        beta : float
            Bowen Ratio
        """

        # Calculate Bowen Ratio 
        beta = (self.air1.c_p / self.air1.lambda_v) * ( 
            (self.air2.theta_a - self.air1.theta_a) / (self.air2.q - self.air1.q ) 
        )

        return beta


    def calc_LE_bowen(self, beta):
    # def calc_LE_bowen(R_n, G, beta):
        """
        Calculates LE using the Bowen Ratio.

        Parameters
        ----------
        R_n : float
            Net radiation [W m-2]

        G : float
            Soil heat flux [W m-2]

        beta : float
            Bowen Ratio

        Returns
        -------
        LE : float
            Latent heat flux [W m-2]

        NOTE: I'm taking the mean R_n and G here. I think this makes more sense
        physically. It's worth noting, though, that slightly different values result
        in the final calculations if these are calculated pixel-wise (and then
        if r_aH is calculated pixel-wise vs. from average H, theta_s).
        """
        # LE = (np.mean(self.radiation.R_n) - np.mean(self.radiation.G)) / (1 + beta)
        LE = (self.radiation.R_n - self.radiation.G) / (1 + beta)

        return LE


    def calc_H_bowen(self, LE, beta):
    # def calc_H_bowen(LE, beta):
        """
        Calculate H using the LE from the Bowen Ratio

        Parameters
        ----------
        LE : float
            Latent heat flux [W m-2]
        
        beta : flaot
            Bowen ratio

        Returns
        -------
        H : float
            Sensible heat flux [W m-2]
        """
        H = beta * LE

        return H
    
    def calc_r_H(self, H_hat):
        """
        Calculate total resistance to heat transfer from sensible heat flux.

        Parameters
        ----------
        H_hat : float
            Landscape-level average sensible heat flux [W m-2].

        Returns
        -------
        r_H : float
            Total resistance to heat transport [s m-1].
        """

        # Calculate aerodynamic resistance
        r_H = self.air2.rho_a * self.air2.c_p * ( (self.surface.theta_s - self.air2.theta_a) / H_hat)

        return r_H


class QuadraticModel(FluxModel):

    def __init__(self, airlayer, surface, radiation):

        super().__init__(airlayer, surface, radiation)


    def run(self, H_0):

        # 1. Calculate Psi terms
        self.Psi_M = self.calc_Psi_M(H_0)
        self.Psi_H = self.calc_Psi_H(H_0)
        # 2. Calculate u_star
        self.u_star = self.calc_ustar(self.Psi_M)
        # 3. Calculate r_aH
        self.r_aH = self.calc_r_aH(self.Psi_M, self.Psi_H)
        # 4. Calculate H
        self.H = self.calc_H(self.surface.theta_s, self.r_aH)
        # 5. Calculate LE
        self.LE = self.calc_LE(self.H)
        # 6. Calculate L
        H_v = self.calc_Hv(self.H, self.LE)
        self.L = self.calc_L(self.u_star, H_v)
        # 7. Calculate zeta
        self.zeta = self.calc_zeta(self.L)


        
    def calc_Psi_M(self, H_hat):
    # def calc_Psi_M_quadratic(u, z, h, rho_a, c_p, theta_a, theta_s, H_hat):
        """
        Calculates the stability correction function for momentum transport based on
        a quadratic solution.

        Parameters
        ----------
        u : float
            Wind speed [m s-1].
        z : float
            Height of wind speed measurement [m].
        h : float
            Canopy height [m].
        rho_a : float
            Air density [kg m-3].
        c_p : float
            Heat capacity of air at constant pressure [J kg-1 K-1].
        theta_a : float
            Potential temperature of air [K].
        theta_s : float
            Potential temperature of the surface [K].
        H_hat : float
            Landscape-level average sensible heat flux [W m-2].

        Returns
        -------
        _type_
            _description_
        """

        # Coefficients
        a = 1
        b = -1 * ( 
            np.log((self.air.z - self.surface.d_0) / self.surface.z_0m) + \
            np.log( (self.air.z - self.surface.d_0) / self.surface.z_0h ) 
        )
        c = ( 
            np.log( (self.air.z - self.surface.d_0) / self.surface.z_0m ) * \
            np.log( (self.air.z - self.surface.d_0) / self.surface.z_0h ) 
        ) - ( (KAPPA ** 2 * self.air.u * self.air.rho_a * self.air.c_p * (self.surface.theta_s - self.air.theta_a) ) / H_hat )

        # Calculate Psi_M
        psi_1,psi_2 = utils.quadratic_formula(a, b, c)

        return psi_2

    def calc_Psi_H(self, H_hat):
        return self.calc_Psi_M(H_hat)



# TODO: Model outputs (self.get_fluxes()) are not consistent. Need to reconfigure 
# functions so that variables are more consistent types.


#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'MIT'
__date__ = 'Mon 22 Aug 22 15:15:04'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           air.py
Compatibility:  Python 3.7.0
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       None

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2022


"""

# TODO: Add dictionary of params for a few veg types (h, w_l, x_lad)

#%% IMPORTS

import numpy as np
import xarray as xr
from . import utils

SIGMA = 5.67036713e-08 

class Surface:
    """
    
    Attributes
    ----------
    cover : str
        Land cover type

    veg : str
        Vegetation type.

    h : float | int
        Canopy height [m]

    T_s : array-like
        Temperature of the surface [K]
    
    Methods
    -------
    calc_d0()
        Calculate the zero-plane displacement height (height at which wind speed goes
        to 0), d_0 [m].
    
    calc_z0m()
        Calculate the aerodynamic roughness length for momemtum transport, z_0m [m].
    
    calc_z0h()
        Calculate the aerodynamic roughness length for heat transport, z_0h [m].

    """

    def __init__(
        self,
        h,
        T_s,
        w_l = 0.01,
        ndvi = None,
        x_lad = 1.,
        b = 10.,
        theta_sun = 60.,
        cover : str = None,     # TODO: Figure out how to force cover and veg to be in a list of options
        veg : str = None,
    ):
        """
        Parameters
        ----------
        h : float
            Height of the canopy [m]
        T_s : float | array-like
            Temperature of the surface [K]
        w_l : float
            Leaf width [m]
        ndvi : float | array-like
            NDVI
        x_lad : float | array-like
            Ratio of horizontal to vertical projected leaf area
        b : float 
            Ratio of z_0m to z_0h. The default is 10.
        theta_sun : float
            Solar zenith angle at time of data collection.
        cover : str
            Land cover type
        veg : str
            Vegetation type
        """

        self.h = h
        self.w_l = w_l
        self.T_s = T_s
        self.ndvi = ndvi
        self.cover = cover
        self.veg = veg

        self.theta_s = self.calc_theta_s()

        self.d_0 = self.calc_d0()
        self.z_0m = self.calc_z0m()
        self.z_0h = self.calc_z0h(b=b)

        if ndvi is not None:
            self.albedo = self.calc_albedo(ndvi)
            self.epsilon_s = self.calc_surf_emissivity(ndvi)
            self.lai = self.calc_lai( k_be = self.calc_k_be(theta_sun=theta_sun, x=x_lad) )
        else:
            self.albedo = 0.28
            self.epsilon_s = 0.98
            self.lai = 1.0

    def describe(self):
        raw = self.__dict__.copy()
        # raw['T_s'] = self.T_s.mean().item()
        raw.update({k : v.mean().item() for k,v in raw.items() if v is not None and not isinstance(v,(float,int,str))})
        return raw

    def calc_theta_s(self):
        return self.T_s


    def calc_d0(self):
        """
        Calculate the zero-plane displacement height (height at which wind speed goes
        to 0), d_0 [m].

        Parameters
        ----------
        h : float
            Canopy height [m].

        Returns
        -------
        d_0 : float
            Zero-plane displacement height [m].

        Reference: Norman et al. (1995).
        """
        d_0 = 0.65 * self.h

        return d_0

    def calc_z0m(self):
        """
        Calculate the aerodynamic roughness length for momemtum transport, z_0m [m].

        Parameters
        ----------
        h : float
            Canopy height [m]

        Returns
        -------
        z_0m : float
            Roughness length for momentum transport [m].

        Reference: Norman et al. (1995).
        """
        z_0m = 0.125 * self.h

        return z_0m


    def calc_z0h(self, b=10.):
        """
        Calculate the aerodynamic roughness length for heat (and vapour) transport,
        z_0h [m].

        Parameters
        ----------
        z_0m : float
            Roughness length for momentum transport [m].

        Returns
        -------
        z_0h : float
            Roughness length for heat transport [m].

        Reference: Bonan (2016), p. 213.
        """
        # TODO: This only applies to vegetation. Should change this?
        z_0h = self.z_0m / b

        return z_0h

    def calc_albedo(self, ndvi):
        """
        Estimate albedo from NDVI.

        Parameters
        ----------
        ndvi : float or array-like
            Normalized-difference vegetation index.

        Returns
        -------
        alpha : float or array-like
            Surface albedo.
        
        Source: GAO, 1995; Mo et al., 2014 in Wang et al. (2019), p. 20, eqs. 13-14.
        (NOTE: There may be a better source out there...)
        """

        # Calculate NDVI reflectance
        sr = (1 + ndvi) / (1 - ndvi)
        # Calculate albedo
        alpha = 0.28 - 0.14 * np.e**( -6.08 / (sr**2) )

        return alpha


    def calc_surf_emissivity(self, ndvi):

        # For 0.131 <= NDVI < 0.608, epsilon_s = 1.0094 + 0.047 * ln(ndvi); else, ndvi = 0.986
        epsilon_s = xr.where( 
            ( (ndvi < 0.608) & (ndvi >= 0.131) ), 
            1.0094 + 0.047 * np.log(ndvi), 
            0.986
        )
        # for NDVI < 0.131, epsilon_s = 0.914 (else leave as calculated above)
        epsilon_s = xr.where( ndvi < 0.131, 0.914, epsilon_s )
        # Set nans
        if not isinstance(ndvi,(float,int)):
            epsilon_s = xr.where( ndvi.isnull(), np.nan, epsilon_s )

        if isinstance(epsilon_s,xr.DataArray):
            utils.set_crs(epsilon_s, self.T_s)

        return epsilon_s


    def correct_T_b(self, LW_IN, T_a, tau, epsilon_s=None):

        if epsilon_s is None:
            epsilon_s = self.epsilon_s

        LW_sens = SIGMA * self.T_s**4
        LW_refl = (1 - epsilon_s) * LW_IN       # Incoming LW radiation reflected by surface
        LW_atm = (1 - tau) * SIGMA * (T_a**4)   # LW radiation emitted by atmosphere btw surface and sensor

        # tes = tau * epsilon_s * SIGMA

        # T_s = ( ( LW_OUT / (tau * epsilon_s * SIGMA) ) - \
        #         ( ((1 - epsilon_s) / (epsilon_s * SIGMA)) * LW_IN ) - \
        #         ( LW_atm / (tau * epsilon_s) ) 
        #     ) ** (1/4)

        if self.T_s.shape != epsilon_s.shape:
            LW_refl = LW_refl.interp(x = self.T_s.x, y = self.T_s.y)


        T_s = ( ( LW_sens - tau * LW_refl - LW_atm ) / (tau * epsilon_s * SIGMA) ) ** (1/4)
        
        self.theta_s = T_s.copy()

        return T_s

    def calc_k_be(self, theta_sun, x=1.):
        """
        Calculate the beam extinction coefficient base on an ellipsoidal leaf 
        inclination distribution function. For spherical LAD, x = 1. For vertical,
        x = 0.

        Parameters
        ----------
        theta_sun : float or array-like
            Solar zenith angle [degrees].
        x : float, optional
            Ratio of horizontal to vertical projected area (i.e. leaf angle distribution, 
            LAD). For spherical LAD, x = 1. For vertical, x = 0.
        
        Returns
        -------
        k_be : float or array-like
            Beam extinction coefficient
        
        Reference: Campbell + Norman, 1998, eq. 15.4
        """

        theta_rad = np.radians(theta_sun)

        k_be = ( np.sqrt( x**2 + np.tan(theta_rad)**2 ) /
                (x + 1.774 * (x + 1.182)**-0.733 ) )

        return k_be

    def calc_vi_scaled(self):
        """
        Calculate scaled vegetation index.

        Parameters
        ----------
        vi : array-like
            Vegetation index

        Returns
        -------
        vi_star : array-like
            Scaled vegetation index.
        
        Reference: Jones and Vaughan, Eq. 7.6
        """
        if isinstance(self.ndvi, (float,int)):
            return self.ndvi
        
        vi_veg = self.ndvi.where(self.ndvi < 1.0).max()
        vi_soil = self.ndvi.where(self.ndvi > 0.0).min()

        vi_star = (self.ndvi - vi_soil) / (vi_veg - vi_soil)

        vi_star = xr.where(vi_star == 1.0, vi_veg, vi_star)
        vi_star = xr.where(vi_star < 0.0, 0.0, vi_star)

        return vi_star
    

    def calc_lai(self, k_be):
        """
        Calculate LAI from scaled vegetation index (typically NDVI).

        Parameters
        ----------
        k_be : float or array-like
            Beam extinction coefficient
        vi_star : float or array-like
            Scaled vegetation index

        Returns
        -------
        lai : float or array-like
            Leaf area index [m m-2]
        
        Reference: Jones and Vaughan, Eq. 7.6

        TODO: Consider moving calc_lai() (and related functions) to ecoflydro.ortho.
        Doesn't make sense for epsilon_s and albedo, but LAI could be useful more
        generally (like NDVI).
        """
        vi_star = self.calc_vi_scaled()

        lai = - (1 / k_be) * np.log(1 - vi_star)

        if isinstance(lai,xr.DataArray):
            utils.set_crs(lai, self.T_s)

        return lai


#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'MIT'
__date__ = 'Thu 13 Oct 22 15:45:11'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           temp.py
Compatibility:  Python 3.7.0
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       None

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2022


"""


# IMPORTS

import numpy as np

# FUNCTIONS


SIGMA = 5.67036713e-08 



def calc_wvc(h_r, T_a):
    """
    Calculate the water vapour content of air [mm Hg]

    Parameters
    ----------
    h_r : float | array-like
        Relative humidity [%]
    T_a : float | array-like
        Air temperature [K]

    Returns
    -------
    wvc : float | array-like
         Water vapour concentration [mm Hg]
        
    SOURCE: 
    """
    h1 = 6.8455e-7
    h2 = -2.7816e-4
    h3 = 6.939e-2
    h4 = 1.5587

    T_aC = T_a - 273.15

    wvc = (h_r/100) * np.exp( 
        h1 * T_aC**3 + h2 * T_aC**2 + h3 * T_aC + h4
    )

    return wvc


def calc_tau(wvc, d):
    """
    Calculate atmospheric transmissivity 

    Parameters
    ----------
    wvc : float | array-like
        Water vapour concentration [mm Hg]
    d : float | array-like
        Distance between surface and sensor [m]

    Returns
    -------
    tau : float | array-like
        Atmospheric transmissivity.
    
    SOURCE: 
    """

    K_atm = 1.9
    a1 = 6.569e-3
    a2 = 0.01262
    b1 = -2.276e-3
    b2 = -6.67e-3

    tau = K_atm * np.exp( -(d**(1/2)) * (a1 + b1 * wvc**(1/2)) ) + \
         (1 - K_atm) * np.exp( -(d**(1/2)) * (a2 + b2 * wvc**(1/2)) )

    return tau




def calc_LW_atm(T_a, tau):
    """
    Calculate the longwave radiation emitted by the atmosphere between the surface and a sensor.

    Parameters
    ----------
    T_a : float | array-like
        Air temperature [K]
    tau : float | array-like
        Atmospheric transmissivity

    Returns
    -------
    LW_atm : float | array-like
        Longwave radiation emitted by the atmosphere between the surface and a sensor [W m-2].
    """

    LW_atm = (1 - tau) * SIGMA * (T_a**4)

    return LW_atm

def calc_LW_refl(LW_IN, epsilon_s):
    """
    Calculate the incoming longwave radiation reflected by a surface.

    Parameters
    ----------
    LW_IN : float | array-like
        Incoming longwave radiation [W m-2]
    epsilon_s : float | array-like
        Surface emissivity

    Returns
    -------
    LW_refl : float | array-like
        Incoming longwave radiation reflected by a surface [W m-2].
    """
    LW_refl = (1 - epsilon_s) * LW_IN

    return LW_refl


def calc_T_s_LW(T_b, LW_IN, T_a, tau, epsilon_s):

    LW_sens = SIGMA * T_b**4
    LW_atm = calc_LW_atm(T_a, tau)              # LW radiation emitted by atmosphere btw surface and sensor
    LW_refl = calc_LW_refl(LW_IN, epsilon_s)    # Incoming LW radiation reflected by surface

    # tes = tau * epsilon_s * SIGMA

    # T_s = ( ( LW_OUT / (tau * epsilon_s * SIGMA) ) - \
    #         ( ((1 - epsilon_s) / (epsilon_s * SIGMA)) * LW_IN ) - \
    #         ( LW_atm / (tau * epsilon_s) ) 
    #     ) ** (1/4)


    T_s = ( ( LW_sens - tau * LW_refl - LW_atm ) / (tau * epsilon_s * SIGMA) ) ** (1/4)
    
    return T_s



def calc_T_s_EMP(T_b, slope, intercept):

    T_s = slope * T_b + intercept

    return T_s


def correct_T_s(T_s, method=None, **kwargs):

    if method == 'LW':
        T_s_corr = calc_T_s_LW(T_b = T_s, **kwargs)
    elif method == 'empirical':
        T_s_corr = calc_T_s_EMP(T_b = T_s, **kwargs)
    else:
        return T_s
    
    return T_s_corr


#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'MIT'
__date__ = 'Wed 24 Aug 22 16:04:30'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           radiation.py
Compatibility:  Python 3.7.0
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       None

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2022


"""


# IMPORTS
import numpy as np
import xarray as xr

# CONSTANTS
SIGMA = 5.67036713e-08 		# Stefan-Boltzmann constant [W m-2 K-4]


def calc_SW_in(SW_in_mean, cos_i):
    """
    Calculate pixel-wise incoming shortwave radiation.

    Parameters
    ----------
    SW_in_mean : float
        Mean shortwave incoming radiation over the flight [W m-2].
        NOTE: May want a different value?--how to make use of the wealth of SW data?
    cos_i : float
        Cosine of the incidence angle of solar radiation.

    Returns
    -------
    SW_in : ndarray
        Incoming shortwave radiation raster [W m-2].
    """

    SW_in = SW_in_mean * cos_i

    return SW_in


def calc_cos_i(slope, aspect, theta_sun, psi_sun):
    """
    Calculate the cosine of the solar incidence angle.

    Parameters
    ----------
    slope : ndarray
        Slope of the surface [radians].
    aspect : ndarray
        Aspect of the surface (orientation of the slope from N) [radians].
    theta_sun : float
        Solar zenith angle [radians].
    psi_sun : float
        Solar azimuth angle [radians].

    Returns
    -------
    [type]
        [description]
    """

    cos_i = ( np.cos(slope) * np.cos(theta_sun) ) + ( 
        np.sin(slope) * np.sin(theta_sun) * np.cos(psi_sun - aspect) 
    )

    return cos_i


def calc_SW_out(SW_in, alpha):

    SW_out = SW_in * alpha

    return SW_out


def calc_albedo(ndvi):
    """
    Estimate albedo from NDVI.

    Parameters
    ----------
    ndvi : float or array-like
        Normalized-difference vegetation index.

    Returns
    -------
    alpha : float or array-like
        Surface albedo.
    
    Source: GAO, 1995; Mo et al., 2014 in Wang et al. (2019), p. 20, eqs. 13-14.
    (NOTE: There may be a better source out there...)
    """

    # Calculate NDVI reflectance
    sr = (1 + ndvi) / (1 - ndvi)
    # Calculate albedo
    alpha = 0.28 - 0.14 * np.e**( -6.08 / (sr**2) )

    return alpha


def calc_emissivity(e_a,T_a, c=0):
    """
    Estimates the atmospheric emissivity for a clear sky.

    Parameters
    ----------
    e_a : float
        Vapor pressure of air [kPa].

    T_a : float
        Temperature of air [K]

    c : float, optional
        Cloud cover fraction [--]. The default is 0 (clear sky).
        
    Returns
    -------
    epsilon_a : float
        Atmospheric emissivity [unitless]

    Reference: Brutsaert (1975), Eq. 11
    Brutsaert, W. (1975) On a derivable formula for long-wave radiation
        from clear skies, Water Resour. Res., 11(5), 742-744,
        htpp://dx.doi.org/10.1029/WR011i005p00742.

    NOTE: Based on tower data, this formula is not great.
    """

    epsilon_ac = 1.24 * ((e_a * 10) / T_a )**(1./7.)

    epsilon_a = calc_cloudy_emissivity(epsilon_ac, c)

    return epsilon_a

def calc_cloudy_emissivity(eps_ac, c):
    """
    Calculates the atmospheric emissivity for a cloudy sky. Note that if c = 0, 
    then eps_a = eps_ac.

    Parameters
    ----------
    eps_ac : float
        Clear-sky atmospheric emissivity [-].
    c : float
        Cloud cover fraction [--].

    Returns
    -------
    eps_a : float
        Cloudy-sky atmospheric emissivity [-].

    Reference: Campbell and Norman (1998), Eq. 10.12
    Campbell, G. S. and J. M. Norman (1998), An Introduction to Environmental
        Biophysics, 2nd Ed., Springer, New York, 286 pp.

    NOTE: Based on tower data, this formula is not great.
    """

    eps_a = ( (1 - 0.84 * c) * eps_ac ) + (0.84 * c)

    return eps_a


def calc_surf_emissivity(ndvi):

    # For 0.131 <= NDVI < 0.608, epsilon_s = 1.0094 + 0.047 * ln(ndvi); else, ndvi = 0.986
    epsilon_s = xr.where( 
        ( (ndvi < 0.608) & (ndvi >= 0.131) ), 
        1.0094 + 0.047 * np.log(ndvi), 
        0.986
    )
    # for NDVI < 0.131, epsilon_s = 0.914 (else leave as calculated above)
    epsilon_s = xr.where( ndvi < 0.131, 0.914, epsilon_s)
    # Set nans
    epsilon_s = xr.where(ndvi.isnull(), np.nan, epsilon_s)
    epsilon_s

    return epsilon_s


def calc_stefboltz(T_K):
    """
    Calculates the total energy radiated by a blackbody according to the
    Stefan-Boltzmann Law.

    Parameters
    ----------
    T_K : float
        Temperature [K].

    Returns
    -------
    M : float
        Emitted blackbody radiance [W m-2]

    """

    M = SIGMA * T_K**4

    return M


def calc_LW(T, epsilon=1):
    """ Add doc strings



    """
    return epsilon * calc_stefboltz(T)


def calc_Rn(SW_IN, SW_OUT, LW_IN, LW_OUT):
    """
    Calculates net radiation from the four components of the surface energy balance.

    Parameters
    ----------
    SW_in : float or array-like
        Incoming shortwave radiation [W m-2].

    SW_out : float or array-like
        Outgoing shortwave radiation [W m-2].
    
    LW_in : float or array-like
        Incoming longwave radiation [W m-2].

    LW_out : float or array-like
        Outgoing longwave radiation [W m-2].

    Returns
    -------
    R_n : float or array-like
        Net radiation [W m-2].
    """
    R_n = SW_IN - SW_OUT + LW_IN - LW_OUT

    return R_n


#-------------------------------------------------------------------------------
#       RADATION
#-------------------------------------------------------------------------------

class Radiation:
    """
    
    Attributes
    ----------
    R_n : array-like
        Net radiation [W m-2]
    
    G : float | array-like
        Ground heat flux [W m-2]
    
    SW_IN : float | array-like
        Shortwave incoming radiation [W m-2]
    
    SW_OUT : float or array-like
        Outgoing shortwave radiation [W m-2].
    
    LW_IN : float or array-like
        Incoming longwave radiation [W m-2].

    LW_OUT : float or array-like
        Outgoing longwave radiation [W m-2].

    Methods
    -------
    calc_Rn()
        Calculates net radiation from the four components of the surface energy balance.
    
    calc_G()
        Calculate ground heat flux.
    
    set_components()
        Set radiation components as attributes.
    """
    def __init__(self, SW_IN, SW_OUT, LW_IN, LW_OUT, R_n = None, G = None, G_params=None, allow_neg=False):
        """

        Parameters
        ----------
        SW_IN : float | array-like
            Shortwave incoming radiation [W m-2]
        SW_OUT : float or array-like
            Outgoing shortwave radiation [W m-2].
        LW_IN : float or array-like
            Incoming longwave radiation [W m-2].
        LW_OUT : float or array-like
            Outgoing longwave radiation [W m-2].
        R_n : float | array-like, optional
            Net radiation [W m-2]. The default is to calculate from components.
        G : float | array-like, optional
            Ground heat flux. The default is to calculate.
        """


        # self.SW_IN = np.array(SW_IN)
        # self.SW_OUT = np.array(SW_OUT)
        # self.LW_IN = np.array(LW_IN)
        # self.LW_OUT = np.array(LW_OUT)

        if R_n is not None:
            self.R_n = R_n
        else:
            self.R_n = self.calc_Rn(SW_IN, SW_OUT, LW_IN, LW_OUT, allow_neg=allow_neg)
        
        if G is not None:
            self.G = G
        elif G_params is not None:
            self.G = self.calc_G(**G_params)
        else:
            self.G = 0.0

        self.set_component_avgs(SW_IN, SW_OUT, LW_IN, LW_OUT)
    

    def describe(self, means=True):
        raw = self.__dict__.copy()
        # raw['R_n'] = np.mean(self.R_n)
        if means:
            raw.update({k : v.mean().item() for k,v in raw.items() if not isinstance(v,(float,int))})

        return raw
            


    @staticmethod
    def calc_Rn(SW_IN, SW_OUT, LW_IN, LW_OUT, allow_neg=False):
        """
        Calculates net radiation from the four components of the surface energy balance.

        Parameters
        ----------
        SW_in : float or array-like
            Incoming shortwave radiation [W m-2].

        SW_out : float or array-like
            Outgoing shortwave radiation [W m-2].
        
        LW_in : float or array-like
            Incoming longwave radiation [W m-2].

        LW_out : float or array-like
            Outgoing longwave radiation [W m-2].

        Returns
        -------
        R_n : float or array-like
            Net radiation [W m-2].
        """
        R_n = SW_IN - SW_OUT + LW_IN - LW_OUT

        # Force negative values to be 0.
        if not isinstance(R_n, float) and not allow_neg:
            R_n = xr.where(R_n < 0, 0.0, R_n)

        return R_n

    def calc_G_ratio(self, t, A=0.2, c=-4.0, B=24.0):
        """
        Calculate the ratio of G / R_n.

        Parameters
        ----------
        t : float or array-like
            Time of day [h].
        A : float, optional
            Amplitude of diurnal fluctuation in G/R_n (max diurnal G/R_n). The 
            default is 0.2.
        c : float, optional
            Phase shift relative to noon [h]. Negative values are in the afternoon. 
            The default is -4.0 (16:00).
        B : float, optional
            Period [h]. The default is 24.

        Returns
        -------
        G_ratio : float or array-like
            Ratio of soil heat flux to net radiation.
        
        Reference
        ---------
        Santanello, J. A., & Friedl, M. A. (2003). Diurnal Covariation in Soil 
        Heat Flux and Net Radiation. Journal of Applied Meteorology and Climatology, 
        42(6), 851–862. https://doi.org/10.1175/1520-0450(2003)042<0851:DCISHF>2.0.CO;2
            
        """
        t = t - 12.0
        G_ratio = A * np.cos(2.0 * np.pi * (t + c) / B)

        return G_ratio


    def calc_G(self, t, A=0.2, c=-4.0, B=24.0):
        """
        Calculate ground heat flux.

        Parameters
        ----------
        t : float or array-like
            Time of day [h].
        A : float, optional
            Amplitude of diurnal fluctuation in G/R_n (max diurnal G/R_n). The 
            default is 0.2.
        c : float, optional
            Phase shift relative to noon [h]. Negative values are in the afternoon. 
            The default is -4.0 (16:00).
        B : float, optional
            Period [h]. The default is 24.

        Returns
        -------
        G : float or array-like
            Ground heat flux [W m-2].

        """
        G_ratio = self.calc_G_ratio(t, A=A, c=c, B=B)
        G = G_ratio * self.R_n

        return G


    def set_components(self, SW_IN, SW_OUT, LW_IN, LW_OUT):

        self.SW_IN = SW_IN
        self.SW_OUT = SW_OUT
        self.LW_IN = LW_IN
        self.LW_OUT = LW_OUT

    def set_component_avgs(self, SW_IN, SW_OUT, LW_IN, LW_OUT):
        
        self.SW_IN_mean = np.mean(SW_IN)
        self.SW_OUT_mean = np.mean(SW_OUT)
        self.LW_IN_mean = np.mean(LW_IN)
        self.LW_OUT_mean = np.mean(LW_OUT)




#!usr/bin/env python
# -*- coding: utf-8 -*-
#––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

__author__ = 'Bryn Morgan'
__contact__ = 'bryn.morgan@geog.ucsb.edu'
__copyright__ = '(c) Bryn Morgan 2022'

__license__ = 'MIT'
__date__ = 'Wed 17 Aug 22 10:21:05'
__version__ = '1.0'
__status__ = 'initial release'
__url__ = ''

"""

Name:           utils.py
Compatibility:  Python 3.10.2
Description:    Description of what program does

URL:            https://

Requires:       list of libraries required

Dev ToDo:       None

AUTHOR:         Bryn Morgan
ORGANIZATION:   University of California, Santa Barbara
Contact:        bryn.morgan@geog.ucsb.edu
Copyright:      (c) Bryn Morgan 2022


"""


# IMPORTS
import numpy as np


def quadratic_formula(a,b,c):
    """
    Return the roots of a quadratic equation, ax^2 + bx + c = 0.

    Parameters
    ----------
    a : float
        Quadratic term.

    b : float
        Linear term

    c : float
        Constant

    Returns
    -------
    root_pos : float
        The "positive" root (-b + sqrt(d)).

    root_neg : float
        The "negative" root (-b - sqrt(d)).
    """

    d = b**2 - 4*a*c
    root_pos = (- b + np.sqrt(d)) / 2*a
    root_neg = (- b - np.sqrt(d)) / 2*a

    return root_pos,root_neg


def calc_lhvap(T_K):
    """
    Calculates the latent heat of vaporization [J kg-1] of water at a given
    temperature.

    Parameters
    ----------
    T_K : float
        Temperature [K].

    Returns
    -------
    lambda_v : float
        Latent heat of vaporization [J kg-1]

    Reference: Allen et al. (1998), Eq. 3-1

    """
    # Calculate latent heat of vaporization at T
    lambda_v = (2.501 - (2.361e-3 * (T_K - 273.15))) * 1e6

    return lambda_v


def calc_gamma(c_p, T_K):

    gamma = c_p / calc_lhvap(T_K)

    return gamma

def set_crs(arr, ref_arr):

    arr.rio.set_crs(ref_arr.rio.crs)



# def calc_stefboltz(T_K):
#     """
#     Calculates the total energy radiated by a blackbody according to the
#     Stefan-Boltzmann Law.

#     Parameters
#     ----------
#     T_K : float
#         Temperature [K].

#     Returns
#     -------
#     M : float
#         Emitted blackbody radiance [W m-2]

#     """

#     M = SIGMA * T_K**4

#     return M


# def calc_emissivity(e_a,T_a):
#     """
#     Estimates the atmospheric emissivity for a clear sky.

#     Parameters
#     ----------
#     e_a : float
#         Vapor pressure of air [kPa].

#     T_a : float
#         Temperature of air [K]

#     Returns
#     -------
#     epsilon_a : float
#         Atmospheric emissivity [unitless]

#     Reference: Brutsaert (1975), Eq. 11
#     Brutsaert, W. (1975) On a derivable formula for long-wave radiation
#         from clear skies, Water Resour. Res., 11(5), 742-744,
#         htpp://dx.doi.org/10.1029/WR011i005p00742.

#     NOTE: Based on tower data, this formula is not great.
#     """

#     epsilon_a = 1.24 * ((e_a * 10) / T_a )**(1./7.)

#     return epsilon_a


